import pathlib
import shlex
from subprocess import DEVNULL, PIPE, Popen

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
MOL_TYPES = ["nucl", "prot"]


def create_blastdb(infasta, db_dir, db_name, mol_type="nucl", verbose=False,
                   parse_seqids=True, hash_index=False, gi_mask=False,
                   mask_data=None, mask_id=None, logfile=None,
                   tax_id=None, tax_id_map=None):
    if verbose:
        stdout = PIPE
    else:
        stdout = DEVNULL

    db_path = db_dir.joinpath(db_name)
    command = f"makeblastdb -in {infasta} -dbtype {mol_type} -out {db_path}"

    if parse_seqids:
        command = " ".join([command, "-parse_seqids"])
    if hash_index:
        command = " ".join([command, "-hash_index"])
    if gi_mask:
        command = " ".join([command, "-gi_mask"])
    if mask_data is not None:
        command = " ".join([command, "-mask_data", str(mask_data)])
    if mask_id is not None:
        command = " ".join([command, "-mask_id", str(mask_id)])
    if logfile is not None:
        command = " ".join([command, "-logile", str(logfile)])
    if tax_id is not None:
        command = " ".join([command,  "-taxid", str(tax_id)])
    if tax_id_map is not None:
        command = " ".join([command, "-taxid_map", str(tax_id_map)])

    split_command = shlex.split(command)
    with Popen(args=split_command, stdout=stdout, stderr=stdout) as process:
        out, errors = process.communicate()

    return db_path


def blastdbcmd(blast_db_path, entries, outfile_path, verbose=0):
    command = (f"blastdbcmd -db {blast_db_path} -out {outfile_path}")

    if isinstance(entries, list):
        command = " ".join([command, "-entry", ",".join([entries])])
    if isinstance(entries, str):
        command = " ".join([command, "-entry", entries])
    elif isinstance(entries, pathlib.Path):
        command = " ".join([command, "-entry_batch", str(entries)])

    command = shlex.split(command)
    with Popen(args=command, stdout=PIPE, stderr=PIPE) as process:
        out, errors = process.communicate()
        if verbose > 0 and out:
            print(out.decode("utf-8"))
        if verbose > 1 and errors:
            print(errors.decode("utf-8"))




from depht.data import GLOBAL_VARIABLES

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
CODING_FEATURE_TYPES = GLOBAL_VARIABLES["sequences"]["feature_types"]


class Contig:
    """Class to manage Bacterial genome contig sequences and attributes
    """
    def __init__(self, record, seq_id):
        self.record = record
        self.seq = record.seq
        self.id = seq_id

        self.genes = list()
        self.gene_ids = list()

        self.model_scores = list()
        self.mask_bits = list()
        self.hhsearch_scores = list()

        self.update_gene_attributes()

    def update_gene_attributes(self):
        """Catalogs the genes in a bacterial sequence contig's record
        """
        if self.record is None:
            return

        self.genes = list()
        self.gene_ids = list()

        for index, feature in enumerate(self.record.features):
            if feature.type not in CODING_FEATURE_TYPES:
                continue

            gene_id = "_".join([self.id, str(index+1)])

            feature.qualifiers["locus_tag"] = [gene_id]
            feature.qualifiers["gene"] = [str(index+1)]

            if feature.type != "CDS":
                continue

            self.genes.append(feature)
            self.gene_ids.append(gene_id)

    def update_model_scores(self, model_scores):
        if len(model_scores) != len(self.genes):
            raise ValueError("Inputted model scores iterable does not match "
                             "the length of the stored gene attributes.")

        self.model_scores = list()
        for model_score in model_scores:
            self.model_scores.append(model_score)

    def update_mask_bits(self, mask_bits):
        if len(mask_bits) != len(self.genes):
            raise ValueError("Inputted gene bit mask iterable does not match "
                             "the length of the stored gene attributes")

        self.mask_bits = list()
        for bit in mask_bits:
            self.mask_bits.append(bit)

    def update_hhsearch_scores(self, hhsearch_scores):
        if len(hhsearch_scores) != len(self.genes):
            raise ValueError("Inputted HHsearch score iterable does not match "
                             "the length of the stored gene attributes")

        self.hhsearch_scores = list()
        for score in hhsearch_scores:
            self.hhsearch_scores.append(score)

    def fill_hhsearch_scores(self):
        if not (self.genes):
            return

        self.hhsearch_scores = [0.0] * len(self.genes)

    def fill_mask_bits(self):
        if not (self.genes):
            return

        self.mask_bits = [1] * len(self.genes)



"""Utility script to build an HHSuite3 HMM database of phams with
conserved phage functions.
"""
import argparse
import pathlib
import sys

from depht.data import GLOBAL_VARIABLES
from depht.functions.multiprocess import CPUS
from depht_train.functions import hhsuitedb

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
NAME = GLOBAL_VARIABLES["phage_homologs"]["essential_name"]


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------
def parse_args(unparsed_args):
    """Parse commandline arguments.

    :param unparsed_args:
    :type unparsed_args: list[str]
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("input_dir", type=pathlib.Path)
    parser.add_argument("output_dir", type=pathlib.Path)

    parser.add_argument("-v", "--verbose",  action="store_true")
    parser.add_argument("-n", "--name", type=str)

    parser.add_argument("-np", "--cpus", type=int, default=CPUS)
    parser.add_argument("-mpi", "--use_mpi", action="store_true")

    args = parser.parse_args(unparsed_args)
    return args


def build_HMM_db(input_dir, output_dir, name=NAME,
                 cores=CPUS, use_mpi=False, verbose=False):
    output_dir.mkdir(exist_ok=True, parents=True)

    hhsuitedb.create_hhsuitedb(input_dir, output_dir, name, cores=cores,
                               use_mpi=use_mpi, verbose=verbose)


def main(unparsed_args):
    """Commandline entrypoint for this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])
    build_HMM_db(args.input_dir, args.output_dir, name=args.name,
                 cores=args.cpus, use_mpi=args.use_mpi,
                 verbose=args.verbose)


if __name__ == "__main__":
    main()



import datetime

DATE = datetime.date.today().strftime("%d-%b-%Y").upper()

GLOBAL_VARIABLES = {
    "sequences": {
        "name": "genes",
        "default_product": "hypothetical protein",
        "annotations": {
            "molecule_type": "DNA",
            "topology": "linear",
            "data_file_division": "PHG",
            "date": DATE,
            "accessions": [],
            "sequence_version": "1",
            "keywords": [],
            "source": "",
            "organism": "",
            "taxonomy": [],
            "comment": [""]},
        "feature_types": ("CDS", "tRNA", "tmRNA")},
    "bacterial_sequences": {
        "name": "bacterial_genes",
        "dir_name": "bacteria"},
    "phage_sequences": {
        "name": "phage_genes",
        "prophage_prefix": "prophi",
        "prophage_delimiter": "-",
        "dir_name": "phage"},
    "phage_homologs": {
        "essential_name": "essential",
        "extended_name": "extended",
        "dir_name": "hhsearch"},
    "shell_db": {
        "fasta_name": "bacterial_genes.fasta",
        "hex_value_name": "bacterial_genes.pbv",
        "dir_name": "mmseqs"},
    "reference_db": {
        "name": "references",
        "dir_name": "blastn"},
    "classifier": {
        "name": "classifier.pkl"},
    "model_storage": {
        "tmp_dir": "tmp",
        "model_dir": "models",
        "home_dir": ".depht"}}

PARAMETERS = {
    "annotation": {
        "min_length": 20000,
        "meta_length": 100000},
    "att_detection": {
        "att_quality_weight": 1,
        "integrase_proximity_weight": 0.6,
        "model_coverage_weight": 0.9,
        "trna_weight": 0,
        "reference_concurrence_weight": 1.5,
        "kmer_size": 5,
        "min_att_score": 2.2,
        "evalue_threshold": 10000,
        "extention_length": 5000,
        "att_sensitivity": 7,
        "blast_sort_key": "bitscore"},
    "phage_homology_search": {
        "evalue": 1E-04,
        "probability": 90,
        "coverage": 50,
        "min_products": 5,
        "min_products_strict": 10},
    "phameration": {
        "cmode": 0,
        "cstep": 1,
        "sens": 8,
        "ident": 0.5,
        "cover": 0.80,
        "evalue": 0.001,
        "min_gcs": 0.7},
    "prophage_prediction": {
        "window": 55}}



"""Function that tries to detect biological file format by sniffing the
first line of a given file."""


def sniff_format(filepath):
    """Read the first line of the file and return its predicted format.

    :param filepath: the path to a file of unknown format
    :type filepath: pathlib.Path
    :return: fmt
    """
    fmt = None

    with open(filepath) as file_sniffer:
        line = file_sniffer.readline()

    # For now, sniffer only knows about FASTA and Genbank flatfile formats
    if line.startswith(">"):
        fmt = "fasta"
    elif line.startswith("LOCUS"):
        fmt = "genbank"

    return fmt



from Bio import (Entrez, SeqIO)


def set_entrez_credentials(tool=None, email=None, api_key=None):
    """Set BioPython Entrez credentials to improve speed and reliability.

    :param tool: Name of the software/tool being used.
    :type tool: str
    :param email: Email contact information for NCBI.
    :type email: str
    :param api_key: Unique NCBI-issued identifier to enhance retrieval speed.
    :type api_key: str
    """
    if tool is not None:
        Entrez.tool = tool
    if email is not None:
        Entrez.email = email
    if api_key is not None:
        Entrez.api_key = api_key


def run_esearch(db="", term="", usehistory="", idtype="", retmax=99999):
    """Search for valid records in NCBI.

    Uses NCBI esearch implemented through BioPython Entrez.

    :param db: Name of the database to search.
    :type db: str
    :param term: Search term.
    :type term: str
    :param usehistory: Indicates if prior searches should be used.
    :type usehistory: str
    :return: Results of the search for each valid record.
    :rtype: dict
    """
    search_handle = Entrez.esearch(db=db, term=term, usehistory=usehistory,
                                   idtype=idtype, retmax=retmax)
    search_record = Entrez.read(search_handle)
    search_handle.close()

    return search_record


def get_records(accession_list, db="nucleotide", rettype="gb", retmode="text"):
    """Retrieve records from NCBI from a list of active accessions.

    Uses NCBI efetch implemented through BioPython Entrez.

    :param accession_list: List of NCBI accessions.
    :type accession_list: list
    :param db: Name of the database to get summaries from (e.g. 'nucleotide').
    :type db: str
    :param rettype: Type of record to retrieve (e.g. 'gb').
    :type rettype: str
    :param retmode: Format of data to retrieve (e.g. 'text').
    :type retmode: str
    :return: List of BioPython SeqRecords generated from GenBank records.
    :rtype: list
    """
    retrieved_records = []
    fetch_query = ",".join(accession_list)
    fetch_handle = Entrez.efetch(db=db, id=fetch_query, rettype=rettype,
                                 retmode=retmode)
    fetch_records = SeqIO.parse(fetch_handle, "genbank")
    for record in fetch_records:
        retrieved_records.append(record)
    fetch_handle.close()

    return retrieved_records


def parse_identifiers_file(identifiers_file):
    """Parses a taxon identifiers file, assuming each ID is on
    a separate line.

    :param identifiers_file: Path to a file containing taxon identifiers.
    :type identifiers_file: pathlib.Path
    :returns: A set of identifiers.
    :rtype: set()
    """
    ids = set()

    with identifiers_file.open(mode="r") as filehandle:
        for line in filehandle:
            ids.add(line.rstrip())

    return ids 


def esearch_taxa(taxon_id, db="nucleotide", idtype="acc", retmax=99999):
    """Searches for genome accession IDs using a taxon identifier.

    :param taxon_id: Taxon identifier to find corresponding accession IDs for.
    :type taxon_id: str
    :param db: Name of the NCBI database to search:
    :type db: str
    :param idtype: Type of identifier to return in the search:
    :type idtype: str
    :param retmax: Number of maximum identifiers to retrieve.
    :type retmax: int
    :returns: Returns specified identifier type related to the taxon ID
    :rtype: list[str]
    """
    taxon_identifier_ref = "".join(["txid", str(taxon_id), "[Orgn]"])

    record = run_esearch(db=db, idtype=idtype, retmax=retmax,
                                term=taxon_identifier_ref)

    return record["IdList"]


def retrieve_genome_records(taxon_ids):
    """Retrieves genome Biopython SeqRecords from a list of taxon NCBI
    identifiers.

    :param taxon_ids: Taxon identifiers to find corresponding genomes from
    :type taxon_ids: list[str]
    :returns: A list of SeqRecord objects retrieved using taxon identifiers.
    :rtype: list[Bio.SeqRecord.SeqRecord]
    """
    genome_accessions = set()
    for taxon_id in taxon_ids:
        taxon_accessions = esearch_taxa(taxon_id)
        for taxon_accession in taxon_accessions:
            genome_accessions.append(taxon_accession)

    genome_accessions = list(genome_accessions)
    records = get_records(genome_accessions)

    return records



"""Program to run DEPhT training pipelines."""

import argparse
import sys
import time

from depht_train.pipelines import (
    benchmark_output, build_reference_db, build_HMM_db, create_model,
    curate_gene_clusters, index_sequences, phamerate,
    pull_sequences, screen_conserved_phams, train_model)

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
PIPELINES = [
    "benchmark_output", "build_reference_db", "build_HMM_db", "create_model",
    "curate_gene_clusters", "index_sequences", "phamerate", "pull_sequences",
    "screen_conserved_phams", "train_model"]

EPILOG = "The 'create_model' pipeline performs all model-building steps. " \
         "The other pipelines can be used in combination for better control " \
         "over the resultant model."


def parse_args(unparsed_args):
    """Use argparse to verify pipeline argument only.

    :param unparsed_args:
    :type unparsed_args:
    """
    parser = argparse.ArgumentParser(description=__doc__, prog="depht_train",
                                     epilog=EPILOG)

    parser.add_argument("pipeline", type=str, choices=PIPELINES,
                        metavar="pipeline",
                        help=f"name of the depht_train pipeline to run "
                             f"[choices: {PIPELINES}]")

    return parser.parse_args(unparsed_args[1:2])


def main():
    """Commandline entrypoint to this module."""
    if len(sys.argv) == 1:
        sys.argv.append("-h")

    unparsed_args = sys.argv

    args = parse_args(unparsed_args)

    start = time.time()

    if args.pipeline == "benchmark_output":
        benchmark_output.main(unparsed_args[1:])
    elif args.pipeline == "build_reference_db":
        build_reference_db.main(unparsed_args[1:])
    elif args.pipeline == "build_HMM_db":
        build_HMM_db.main(unparsed_args[1:])
    elif args.pipeline == "create_model":
        create_model.main(unparsed_args[1:])
    elif args.pipeline == "curate_gene_clusters":
        curate_gene_clusters.main(unparsed_args[1:])
    elif args.pipeline == "index_sequences":
        index_sequences.main(unparsed_args[1:])
    elif args.pipeline == "phamerate":
        phamerate.main(unparsed_args[1:])
    elif args.pipeline == "pull_sequences":
        pull_sequences.main(unparsed_args[1:])
    elif args.pipeline == "screen_conserved_phams":
        screen_conserved_phams.main(unparsed_args[1:])
    elif args.pipeline == "train_model":
        train_model.main(unparsed_args[1:])
    else:
        raise NotImplementedError(f"unknown depht_train pipeline "
                                  f"{args.pipeline}")

    stop = time.time()

    print("\n\nPipeline completed.\nTime elapsed {:.2f}s".format(stop - start))



"""Utility script to perform pham assembly with MMseqs2.
"""
import argparse
import json
import pathlib
import shutil
import sys

from depht.functions import mmseqs
from depht.functions.fasta import parse_fasta
from depht_train.classes.database import Database, Pham


def parse_args(unparsed_args):
    """Parse commandline arguments.

    :param unparsed_args:
    :type unparsed_args: list[str]
    """
    p = argparse.ArgumentParser(description=__doc__)

    p.add_argument("fasta", type=pathlib.Path,
                   help="path to the FASTA file containing genes to phamerate")
    p.add_argument("outdir", type=pathlib.Path,
                   help="directory where file I/O can occur")
    p.add_argument("param_file", type=pathlib.Path,
                   help="path to the JSON file of MMseqs2 parameters to use")

    return p.parse_args(unparsed_args)


def parse_param_file(param_file):
    """Attempts to parse a given parameter JSON file to two
    tuples of first-iter and second-iter parameter sets for MMseqs2.

    :param param_file: the JSON-formatted parameter file
    :type param_file: pathlib.Path
    :return: first_iter, second_iter
    """
    with param_file.open("r") as param_handle:
        param_dict = json.load(param_handle)

    return parse_param_dict(param_dict)


def parse_param_dict(param_dict):
    """
    Attempts to parse a given parameter dictionary to two
    tuples of first-iter and second-iter parameter sets for MMseqs2.

    :param param_dict: the formatted parameter dictionary
    :type param_dict: dict
    :return: first_iter, second_iter
    """
    first_iter_dict = param_dict.get("first_iteration")
    a = first_iter_dict.get("--cluster-mode")       # Define clustermode
    b = first_iter_dict.get("--cluster-steps")      # Define clustersteps
    c = first_iter_dict.get("-s")                   # Define sensitivity
    d = first_iter_dict.get("--min-seq-id")         # Define min-seq-id
    e = first_iter_dict.get("-c")                   # Define coverage
    f = first_iter_dict.get("-e")                   # Define e-value
    first_iter = (a, b, c, d, e, f)

    second_iter_dict = param_dict.get("second_iteration")
    if second_iter_dict is None:
        second_iter = None
    else:
        g = second_iter_dict.get("--min-seq-id")        # Define min-seq-id
        h = second_iter_dict.get("-c")                  # Define coverage
        j = second_iter_dict.get("-e")                  # Define e-value
        second_iter = (g, h, j)

    return first_iter, second_iter


def phamerate(sequence_db, db, tmpdir, first_iter, second_iter=None):
    """

    :param sequence_db: path to the MMseqs2 sequence database
    :type sequence_db: pathlib.Path
    :param db: database of geneids/translations being phamerated
    :type db: Database
    :param tmpdir:
    :type tmpdir: pathlib.Path
    :param first_iter:
    :type first_iter: tuple
    :param second_iter:
    :type second_iter: tuple
    :return:
    """
    # Refresh tmpdir
    if not tmpdir.is_dir():
        tmpdir.mkdir(parents=False)

    temp_phams = list()

    # First iteration paths
    cluster_db = tmpdir.joinpath("clusterDB")
    seqfile_db = tmpdir.joinpath("seqfileDB")
    first_iter_out = tmpdir.joinpath("first_iter_out.txt")

    # Expand first iteration parameter set
    cmode, cstep, sens, ident, cover, evalue = first_iter

    # Cluster database using given parameters
    mmseqs.mmseqs_cluster(sequence_db, cluster_db, tmpdir, cmode, cstep, sens,
                          ident, cover, evalue)
    mmseqs.mmseqs_createseqfiledb(sequence_db, cluster_db, seqfile_db)
    mmseqs.mmseqs_result2flat(sequence_db, sequence_db, seqfile_db,
                              first_iter_out)
    first_iter_phams = mmseqs.parse_mmseqs(first_iter_out)

    if second_iter:
        # Second iteration paths
        profile_db = tmpdir.joinpath("profileDB")
        consensus_db = tmpdir.joinpath("consensusDB")
        align_db = tmpdir.joinpath("alignDB")
        result_db = tmpdir.joinpath("resultDB")
        hmm_seqfile_db = tmpdir.joinpath("hmmSeqfileDB")
        second_iter_out = tmpdir.joinpath("second_iter_out.txt")

        # Expand second iteration parameter set
        ident, cover, evalue = second_iter

        # Cluster profiles using given parameters
        mmseqs.mmseqs_result2profile(sequence_db, cluster_db, profile_db)
        mmseqs.mmseqs_profile2consensus(profile_db, consensus_db)
        mmseqs.mmseqs_search(profile_db, consensus_db, align_db, tmpdir,
                             ident, cover, evalue)
        mmseqs.mmseqs_clust(consensus_db, align_db, result_db)
        mmseqs.mmseqs_createseqfiledb(sequence_db, result_db, hmm_seqfile_db)
        mmseqs.mmseqs_result2flat(sequence_db, sequence_db, hmm_seqfile_db,
                                  second_iter_out)
        second_iter_phams = mmseqs.parse_mmseqs(second_iter_out)

        lookup = dict()
        for phamid, pham_geneids in first_iter_phams.items():
            for pham_geneid in pham_geneids:
                lookup[pham_geneid] = phamid

        for source_id, source_geneids in second_iter_phams.items():
            all_pham_geneids = set()
            for source_geneid in source_geneids:
                target_id = lookup[source_geneid]
                for target_geneid in first_iter_phams[target_id]:
                    all_pham_geneids.add(target_geneid)
            all_pham_geneids = list(all_pham_geneids)
            all_pham_translations = [db.get_translation_from_geneid(x) for x in
                                     all_pham_geneids]
            temp_phams.append(Pham(all_pham_geneids, all_pham_translations))

    else:
        for pham_id, all_pham_geneids in first_iter_phams.items():
            all_pham_translations = [db.get_translation_from_geneid(x) for x in
                                     all_pham_geneids]
            temp_phams.append(Pham(all_pham_geneids, all_pham_translations))

    return temp_phams


def execute_phamerate_pipeline(fasta, outdir, first_iter, second_iter=None):
    # Read fasta file and build database instance
    geneids, translations = parse_fasta(fasta)
    if not geneids:
        return None 

    tempdir = outdir.joinpath("temp")

    # Make sure outdir and tmpdir exist
    if not outdir.is_dir():
        outdir.mkdir(parents=True)
    if not tempdir.is_dir():
        tempdir.mkdir(parents=False)

    input_db = Database(geneids, translations)

    # Create mmseqsdb
    nr_fasta = outdir.joinpath("genes.fasta")
    with nr_fasta.open("w") as fh:
        fh.write(repr(input_db))
    mmseqsdb = tempdir.joinpath("sequenceDB")
    mmseqs.mmseqs_createdb(nr_fasta, mmseqsdb)

    # Phamerate!
    phams = phamerate(mmseqsdb, input_db, tempdir, first_iter,
                      second_iter)

    # Write fasta file for each pham
    for i, pham in enumerate(phams):
        pham_fasta = outdir.joinpath(f"pham_{i}.fasta")
        pham_translations = pham.get_translations()

        for translation in pham_translations:
            for geneid in input_db.get_geneids_from_translation(translation):
                try:
                    pham.add_gene(geneid, translation)
                except ValueError:
                    pass    # geneid already in

        with pham_fasta.open("w") as fh:
            fh.write(str(pham))

    # Clean up temporary directory
    shutil.rmtree(tempdir)

    # Clean up temporary mulitple sequence fasta file
    if nr_fasta.is_file():
        nr_fasta.unlink()

    return phams 


def main(unparsed_args=None):
    """Commandline entrypoint to this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])

    first_iter_params, second_iter_params = parse_param_file(args.param_file)

    execute_phamerate_pipeline(args.fasta, args.outdir,
                               first_iter_params, second_iter_params)


if __name__ == "__main__":
    main()



import configparser
import sys
from configparser import Error as ConfigError

DEFAULT_CONFIG_SECTIONS = {"ncbi": {"api_key", "email", "tool"}}


def setup_section(keys, value):
    section = {}
    for key in keys:
        section[key] = value
    return section


def default_parser(null_value):
    """Constructs complete config with empty values."""
    # Need to allow no value if the null value is None.
    null_parser = configparser.ConfigParser(allow_no_value=True)
    config_sections = DEFAULT_CONFIG_SECTIONS

    for header in config_sections.keys():
        config_section = config_sections[header]
        for subheader in config_section.keys():
            config_section[subheader] = null_value

    return null_parser


def parse_config(config_path, parser=None):
    """Get parameters from config file."""
    if parser is None:
        parser = configparser.ConfigParser()

    try:
        parser.read(config_path)
    except ConfigError:
        print("Unable to parse config file.")
        sys.exit(1)
    else:
        return parser


def build_complete_config(config_path):
    "Buid a complete config object by merging user-supplied and default config"
    parser = default_parser(None)

    if config_path is not None:
        parser = parse_config(config_path, parser)

    return parser



"""Pipeline to compile the complete set of databases and data structures
required to run DEPhT from bacterial and phage sequences provided.
"""
import argparse
import json
import pathlib
import shutil
import sys

from Bio import SeqIO
from phamclust.__main__ import get_clusters

from depht.__main__ import MODEL_DIR
from depht.data import GLOBAL_VARIABLES
from depht.functions.annotation import (annotate_record,
                                        cleanup_flatfile_records)
from depht.functions.fasta import parse_fasta
from depht.functions.multiprocess import CPUS, parallelize
from depht.functions.sniff_format import sniff_format
from depht_train.data import PARAMETERS
from depht_train.functions import fileio
from depht_train.pipelines.build_HMM_db import build_HMM_db
from depht_train.pipelines.build_reference_db import build_reference_db
from depht_train.pipelines.curate_gene_clusters import (
    curate_gene_clusters)
from depht_train.pipelines.index_sequences import index_sequences
from depht_train.pipelines.phamerate import (execute_phamerate_pipeline,
                                             parse_param_dict)
from depht_train.pipelines.screen_conserved_phams import (
    screen_conserved_phams)
from depht_train.pipelines.train_model import train_model


def parse_args(unparsed_args):
    """Function to parse command line arguments for running the create_model
    pipeline.

    :param unparsed_args: List of command line arguments
    :type unparsed_args: list
    """
    parser = argparse.ArgumentParser()

    parser.add_argument("name", type=str,
                        help="a name for the newly created model")
    parser.add_argument("phages", type=pathlib.Path,
                        help="directory containing phage sequences to train "
                             "with")
    parser.add_argument("bacteria", type=pathlib.Path,
                        help="directory containing bacterial sequences to "
                             "train with")

    parser.add_argument("-v", "--verbose", action="store_true",
                        help="print progress messages to console")
    parser.add_argument("-c", "--cpus", type=int, default=CPUS, metavar="",
                        help=f"number of CPU cores to use [default: {CPUS}]")
    parser.add_argument("-f", "--force", action="store_true",
                        help="overwrite existing model with the same name")

    parser.add_argument("--config", type=pathlib.Path, metavar="",
                        help="configuration file to customize pipeline "
                             "parameters")
    parser.add_argument("--bacteria-clusters", type=pathlib.Path, metavar="",
                        help="a CSV file with manually determined clusters "
                             "for the input bacterial sequences")
    parser.add_argument("--prophage-coords", type=pathlib.Path, metavar="",
                        help="a CSV file with the coordinates of any known "
                             "prophages in the input bacterial sequences")

    return parser.parse_args(unparsed_args)


def load_config(config_file):
    """Function to load the master configuration file for the create_model
    pipeline and to verify its structure/contents.
    """
    config = PARAMETERS
    if config_file is None:
        return config
    elif not config_file.is_file():
        return config

    with open(config_file, "r") as filehandle:
        input_config = json.load(filehandle)

    config.update(input_config)
    return config


def create_model(name, phages, bacteria, bact_clusters=None,
                 prophage_coords=None, config=None, force=False,
                 verbose=False, cpus=1):
    """Function to create a DEPhT model from bacterial and phage sequences.

    :param name: Name given to the created model.
    :type name: str
    :param phages: Directory of phage sequence files.
    :type phages: pathlib.Path
    :param bacteria: Directory of bacterial sequence files.
    :type bacteria: pathlib.Path
    :param verbose: Boolean to toggle pipeline verbosity.
    :type verbose: bool
    :param config: Config file to overwrite pipeline parameters.
    :type config: pathlib.Path
    :param force: Allows pipeline to overwrite already existing models:
    :type force: bool
    :param cpus: Number of CPUs to use during the pipeline.
    :type cpus: int
    :param bact_clusters: File describing bacteria clade membership
    :type bact_clusters: pathlib.Path
    :param prophage_coords: File describing known prophage content in bacteria
    :type prophage_coords: pathlib.Path
    """
    # Load master configuration file, which contains paths for
    # configuration of sub-pipelines
    config = load_config(config)
    if config is None:
        print(f"Specified configuration file at {config} "
              "is incorrectly formatted or contains invalid information.\n"
              "Please check the formatting/contents of this file before "
              "continuing.")
        return

    # Proactively create local model directory structure
    dir_map = create_model_structure(name, force=force)
    if dir_map is None:
        print("There already exists a DEPHT model with the name "
              f"'{name}'.\n"
              "Please rename your model or use the -f flag to forcibly "
              "overwrite the already existing model.")
        return

    # Collect bacterial sequences and write fasta and genbank files
    if verbose:
        print("Collecting/annotating bacterial sequences...")
    bacterial_data_tuple = collect_sequences(
        bacteria, dir_map["bacterial_tmp"],
        GLOBAL_VARIABLES["bacterial_sequences"]["name"],
        config["bacterial_sequences"],
        annotate=True, verbose=verbose, cpus=cpus,
        cluster_table=bact_clusters)

    if bacterial_data_tuple[2] == 0:
        print("Could not recognize any bacterial sequence files "
              "in the specified directory.\n"
              "Please check the formatting and contents of your files.")
        shutil.rmtree(dir_map["model_dir"])
        return
    elif bacterial_data_tuple[8] is None:
        print("Error occured while phamerating bacterial sequence data.\n"
              "Please check your local MMseqs2 installation.")
        shutil.rmtree(dir_map["model_dir"])
        return

    # Collect phage sequences and write fasta and genbank files
    if verbose:
        print("Collecting/annotating phage sequences...")
    phage_data_tuple = collect_sequences(
        phages, dir_map["phage_tmp"],
        GLOBAL_VARIABLES["phage_sequences"]["name"],
        config["phage_sequences"],
        verbose=verbose, cpus=cpus, cluster=False,
        singletons=True)

    if phage_data_tuple[2] == 0:
        print("Could not recognize any phage sequence files "
              "in the specified directory.\n"
              "Please check the formatting and contents of your files.")
        shutil.rmtree(dir_map["model_dir"])
        return
    if phage_data_tuple[8] is None:
        print("Error occuerd while phamerating bacterial sequence data.\n"
              "Please check your local MMseqs2 installation.")
        shutil.rmtree(dir_map["model_dir"])
        return

    if verbose:
        print("Training phage/bacterial classifier...")
    train_model(name, phage_data_tuple[1], bacterial_data_tuple[1],
                cpus=cpus, prophages=prophage_coords)

    if verbose:
        print("\nBuilding bacterial reference database...")
    build_reference_db(bacterial_data_tuple[0], dir_map["reference_db_dir"])

    if not bacterial_data_tuple[7]:
        print(
          "WARNING: not enough bacterial sequences to define proper clades "
          "for pangenomic analyses.\n"
          "\tUnable to create a shell genome content database.\n"
          "\tConsider populating your dataset with more bacterial sequences "
          "and/or sequences more closely related.\n")
    else:
        if verbose:
            print("\nBuilding shell genome content database...")
        rep_threshold = config["shell_db"]["rep_threshold"]
        screen_conserved_phams(
                           bacterial_data_tuple[5], dir_map["shell_db_dir"],
                           bacterial_data_tuple[4], bacterial_data_tuple[6],
                           rep_threshold=rep_threshold)

    # Moves bacterial gene product fasta file into the shell genome database
    bacterial_data_tuple[3].replace(dir_map["shell_db_dir"].joinpath(
                                                bacterial_data_tuple[3].name))

    if verbose:
        print("\nBuilding phage homolog profile database...")
    create_phage_homologs_db(phage_data_tuple, dir_map["phage_homologs_dir"],
                             config, dir_map["phage_homologs_tmp"],
                             verbose=verbose, cpus=cpus)

    # Cleans up after the pipeline
    shutil.rmtree(dir_map["tmp_dir"])


def create_model_structure(model_name, force=False):
    """Function to create a DEPhT model's directory structure.

    :param model_name: Name given to the created model.
    :type model_name: str
    :param force: Allows pipeline to overwrite already existing models.
    :type force: bool
    """
    dir_map = dict()

    # Create local model directory
    model_dir = MODEL_DIR.joinpath(model_name)
    dir_map["model_dir"] = model_dir
    if model_dir.is_dir():
        if force:
            shutil.rmtree(model_dir)
        else:
            return None

    model_dir.mkdir()

    # Create temporary working directory
    tmp_dir = model_dir.joinpath("tmp")
    dir_map["tmp_dir"] = tmp_dir
    tmp_dir.mkdir()

    # Create temporary bacterial sequence directory
    bacterial_tmp = tmp_dir.joinpath("bacteria")
    dir_map["bacterial_tmp"] = bacterial_tmp
    bacterial_tmp.mkdir()

    # Create temporary phage sequence directory
    phage_tmp = tmp_dir.joinpath("phage")
    dir_map["phage_tmp"] = phage_tmp
    phage_tmp.mkdir()

    # Create model reference blast database directory
    reference_db_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["reference_db"]["dir_name"])
    dir_map["reference_db_dir"] = reference_db_dir
    reference_db_dir.mkdir()

    # Create model shell database directory and temporary directory
    shell_db_tmp = tmp_dir.joinpath(
                                GLOBAL_VARIABLES["shell_db"]["dir_name"])
    dir_map["shell_db_tmp"] = shell_db_tmp
    shell_db_tmp.mkdir()
    shell_db_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["shell_db"]["dir_name"])
    dir_map["shell_db_dir"] = shell_db_dir
    shell_db_dir.mkdir()

    # Create model phage homolog HMM profile directory and temporary directory
    phage_homologs_tmp = tmp_dir.joinpath(
                                GLOBAL_VARIABLES["phage_homologs"]["dir_name"])
    dir_map["phage_homologs_tmp"] = phage_homologs_tmp
    phage_homologs_tmp.mkdir()
    phage_homologs_dir = model_dir.joinpath(
                                GLOBAL_VARIABLES["phage_homologs"]["dir_name"])
    dir_map["phage_homologs_dir"] = phage_homologs_dir
    phage_homologs_dir.mkdir()
    return dir_map


def collect_sequences(sequences, output_dir, name, config,
                      annotate=False, verbose=False, cpus=1,
                      singletons=False, cluster=True, cluster_table=None):
    """Function to annotate, phamerate gene products, and cluster input
    sequences.

    :param sequences: Directory of sequence files.
    :type sequences: pathlib.Path
    :param output_dir: Directory to write sequence files to.
    :type output_dir: pathlib.Path
    :param name: Name to give to the stem of the created sequence files.
    :type name: str
    :param config: Config dictionary that contains pipeline parameters
    :type config: dict
    :param annotate: Controls whether all sequence files are auto-annotated
    :type annotate: bool
    :param verbose: Boolean to toggle pipeline verbosity.
    :type verbose: bool
    :param cpus:
    :type cpus: int
    :param singletons: Controls whether singleton clusters are retained.
    :type singletons: bool
    :param cluster: Controls whether sequences are clustered.
    :type cluster: bool
    :param cluster_table: File describing sequence clade membership
    :type cluster_table: pathlib.Path
    """
    fasta_sequences, gb_sequences, seq_count = clean_sequences(
                                    sequences, output_dir, annotate=annotate,
                                    verbose=verbose, cpus=cpus)

    if seq_count == 0:
        return (fasta_sequences, gb_sequences, seq_count)

    fasta_file, index_file, cluster_file = index_sequences(
                                    gb_sequences, output_dir,
                                    name=name, cluster_table=cluster_table)

    gene_clusters_dir = output_dir.joinpath("phams")

    if verbose:
        print("...clustering gene product sequences...")
    phamerate_config = config["phameration"]
    first_iter_params, second_iter_params = parse_param_dict(phamerate_config)
    phams = execute_phamerate_pipeline(fasta_file, gene_clusters_dir,
                                       first_iter_params, second_iter_params)

    if not phams:
        return (fasta_sequences, gb_sequences, seq_count, fasta_file,
                index_file, gene_clusters_dir, cluster_file, None, phams)

    cluster_map = None
    if cluster_file is None and cluster:
        if verbose:
            print("...clustering genomes based on shared gene content...")
        cluster_file, cluster_map = create_cluster_schema(
                                            index_file, gene_clusters_dir,
                                            output_dir, name, config,
                                            singletons=singletons)

    return (fasta_sequences, gb_sequences, seq_count, fasta_file,
            index_file, gene_clusters_dir, cluster_file, cluster_map, phams)


def clean_sequences(input_dir, output_dir, annotate=False, verbose=False,
                    cpus=CPUS, trna=False):
    """Formats sequence files, and annotates them as applicable.

    :param input_dir: Directory of sequence files
    :type input_dir: pathlib.Path
    :param output_dir: Directory to write sequence files to.
    :type output_dir: pathlib.Path
    :param annotate: Controls whether sequence files are auto-annotated.
    :type annotate: bool
    :param verbose: Boolean to toggle pipeline verbosity.
    :type verbose: bool
    :param cpus: Number of CPUs to use during the pipeline.
    :type cpus: int
    :param trna: Controls whether sequence files are annotated with tRNAs.
    :type trna: bool
    """
    fasta_dir = output_dir.joinpath("fasta")
    fasta_dir.mkdir()

    gb_dir = output_dir.joinpath("gb")
    gb_dir.mkdir()

    work_items = list()
    for input_file in input_dir.iterdir():
        work_items.append((input_file, output_dir, fasta_dir, gb_dir,
                           annotate, trna))

    seq_count = sum(parallelize(work_items, cpus, clean_sequence,
                                verbose=verbose))

    return fasta_dir, gb_dir, seq_count


def clean_sequence(input_file, output_dir, fasta_dir, gb_dir, annotate=False,
                   trna=False):
    """Process function to format a sequence file and annotates it as
    applicable.

    :param input_file: Directory of sequence files
    :type input_file: pathlib.Path
    :param output_dir:
    :type output_dir:
    :param fasta_dir: Directory to write the fasta-formatted sequence file to.
    :type fasta_dir: pathlib.Path
    :param gb_dir: Directory to write the gb-formatted sequence file to.
    :type gb_dir: pathlib.Path
    :param annotate: Controls whether sequence files are auto-annotated.
    :type annotate: bool
    :param trna: Controls whether sequence files are annotated with tRNAs.
    :type trna: bool
    """
    file_fmt = sniff_format(input_file)
    if file_fmt not in ("fasta", "genbank"):
        return 0

    records = [record for record in SeqIO.parse(input_file, file_fmt)]

    if file_fmt == "fasta" or annotate:
        for record in records:
            record.features = list()
            annotate_record(record, output_dir, trna=trna)

    cleanup_flatfile_records(records)

    fasta_file = fasta_dir.joinpath(".".join([input_file.stem, "fasta"]))
    SeqIO.write(records, fasta_file, "fasta")

    gb_file = gb_dir.joinpath(".".join([input_file.stem, "gb"]))
    SeqIO.write(records, gb_file, "gb")

    return 1


def create_phage_homologs_db(phage_data_tuple, output_dir, config, tmp_dir,
                             cpus=CPUS, verbose=False):
    """

    :param phage_data_tuple:
    :param output_dir:
    :param config:
    :param tmp_dir:
    :param cpus:
    :param verbose:
    :return:
    """
    min_hmm_count = config["phage_homologs"]["min_hmm_count"]

    if verbose:
        print("...curating for essential phage protein clusters...")
    curated_clusters_dir = tmp_dir.joinpath("curated_phams")
    curated_clusters_dir.mkdir()
    accepted_functions = config["phage_homologs"][
                                "essential_annotations"]["LIKE"]
    ignored_functions = config["phage_homologs"][
                                "essential_annotations"]["NOT LIKE"]
    curate_gene_clusters(phage_data_tuple[5], phage_data_tuple[4],
                         curated_clusters_dir,
                         accepted_functions, ignored_functions,
                         verbose=verbose, cores=cpus,
                         min_hmm_count=min_hmm_count)

    if not [cluster for cluster in curated_clusters_dir.iterdir()]:
        print("WARNING: not enough congruent data to create "
              "essential phage profiles.\n"
              "\tUnable to create a sequence profile database.\n"
              "\tConsider adding more phage sequences and/or "
              "curating the quality of your sequence annotations.\n")
    else:
        if verbose:
            print(
                "...creating database of essential phage protein profiles...")
        build_HMM_db(curated_clusters_dir, output_dir, name="essential")

    if verbose:
        print("...curating for accessory phage protein clusters...")
    extended_curated_clusters_dir = tmp_dir.joinpath("extended_curated_phams")
    extended_curated_clusters_dir.mkdir()
    accepted_functions = config["phage_homologs"][
                                "extended_annotations"]["LIKE"]
    ignored_functions = config["phage_homologs"][
                               "extended_annotations"]["NOT LIKE"]
    curate_gene_clusters(phage_data_tuple[5], phage_data_tuple[4],
                         extended_curated_clusters_dir,
                         accepted_functions, ignored_functions,
                         verbose=verbose,
                         accept_all=True, cores=cpus,
                         min_hmm_count=min_hmm_count)

    if not [cluster for cluster in extended_curated_clusters_dir.iterdir()]:
        print("WARNING: not enough congruent data to create "
              "accessory phage protein profiles.\n"
              "\tUnable to create a sequence profile database.\n"
              "\tConsider adding more phage sequences and/or "
              "curating the quality of your sequence annotations.\n")
    else:
        if verbose:
            print(
               "...creating database of accessory phage protein profiles...")
        build_HMM_db(extended_curated_clusters_dir, output_dir,
                     name="extended")


def create_cluster_schema(index_file, gene_clusters_dir, tmp_dir, name,
                          config, singletons=False):
    """Create the cluster schema for a group of sequences.

    :param index_file: Tsv table index of sequence gene products
    :type index_file: pathlib.Path
    :param gene_clusters_dir: Directory of pham (clustered gene product) fastas
    :type gene_clusters_dir: pathlib.Path
    :param tmp_dir: Directory to put pipeline temporary data.
    :type  tmp_dir: pathlib.Path
    :param name: Name to give to the stem of the created sequence files.
    :type name: str
    :param config: Config  dictionary that contains pipeline parameters
    :type config: dict
    :param singletons: Controls whether singleton clusters are retained.
    :type singletons: bool
    """
    gene_index = fileio.read_gene_index_file(index_file)

    genome_pham_tuples = list()
    for pham_file in gene_clusters_dir.iterdir():
        if pham_file.suffix != ".fasta":
            continue

        pham = pham_file.stem
        pham_headers, pham_sequences = parse_fasta(pham_file)

        for header in pham_headers:
            parent_id = gene_index[header]["parent"]
            genome_pham_tuples.append([parent_id, pham])

    genome_pham_tsv_file = tmp_dir.joinpath(".".join([name, "tsv"]))
    with open(genome_pham_tsv_file, "w") as filehandle:
        for genome_pham_tuple in genome_pham_tuples:
            line = "\t".join(genome_pham_tuple)
            filehandle.write(f"{line}\n")

    clusters = get_clusters(genome_pham_tsv_file,
                            config["pham_clust"]["gcs_threshold"])

    clustered_ids = list()
    cluster_map = dict()
    for i, cluster_matrix in enumerate(clusters):
        if len(cluster_matrix.node_names) <= 1 and not singletons:
            continue

        clustered_ids.append(cluster_matrix.node_names)
        cluster_map[i+1] = cluster_matrix.node_names

    cluster_file = tmp_dir.joinpath(".".join([name, "ci"]))
    fileio.write_cluster_file(clustered_ids, cluster_file)

    return cluster_file, cluster_map


def main(unparsed_args=None):
    """Main function for the command-line interface of the create_model
    pipeline.
    """
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])

    create_model(args.name, args.phages, args.bacteria,
                 bact_clusters=args.bacteria_clusters,
                 prophage_coords=args.prophage_coords,
                 config=args.config, verbose=args.verbose,
                 force=args.force, cpus=args.cpus)


if __name__ == "__main__":
    main()



import pandas
import pretty_html_table
import bokeh
from bokeh import plotting
from bokeh.core.properties import value
from bokeh.embed import file_html
from bokeh.models import HoverTool, Range1d, Title
from bokeh.resources import CDN
from dna_features_viewer import CircularGraphicRecord
from matplotlib import pyplot

from depht.classes.file_translator import (
    CircularSourceFeatureTranslator, LinearFeatureTranslator,
    DEFAULT_FONT_FAMILY)

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEFAULT_NAME = "cartoon"


# BOKEH/DNAFEATUREVIEWER FUNCTIONS
# -----------------------------------------------------------------------------
def draw_complete_diagram(outdir, contigs, prophages, tmp_dir,
                          name=DEFAULT_NAME):
    """
    Plots a circular genome diagram of the host bacterial record(s), linear
    genome diagrams for the predicted prophage record(s), and a table
    containing metadata assocated with their relationship.

    :param outdir: File directory the graphics will be written to
    :type outdir: pathib.Path
    :type contigs: Auto-annotated bacterial sequence contigs
    :type contigs: list
    :param prophage_paths: Filepaths of the associated prophages
    :type prophage_paths: list
    :param tmp_dir: Path to place modified file resources
    :type tmp_dir: pathib.Path
    """
    # Scrub each bacterial record contig of non-source features
    # and write to file
    scrub_host_records(contigs)

    # Scrape metadata from each genome contig record
    host_metadata = scrape_and_tabulate_host_metadata(contigs)

    # Create a circular image resource for each genome contig record
    host_graphic_resources = create_host_graphic_resources(contigs,
                                                           tmp_dir)

    contig_prophage_map = {}
    for prophage in prophages:
        contig_prophages = contig_prophage_map.get(prophage.parent_record.id,
                                                   list())
        contig_prophages.append(prophage)
        contig_prophage_map[prophage.parent_record.id] = contig_prophages

    filepath = outdir.joinpath(f"{name}.html")

    filehandle = open(filepath, "w")
    for i in range(len(contigs)):
        contig = contigs[i]
        if not contig.features:
            continue

        graph = host_graphic_resources[i]
        table = host_metadata[i]

        filehandle.write(graph)

        table_html = ("".join([table, "<br>"]))
        filehandle.write(table_html)

        contig_prophages = contig_prophage_map.get(contig.id, list())

        # Embed prophage plots
        embedded_prophage_plots = embed_prophage_genomes(contig_prophages)

        for prophage_plot in embedded_prophage_plots:
            filehandle.write(prophage_plot)
            filehandle.write("<br><br>")

    filehandle.close()

    bokeh.io.reset_output()


def scrub_host_records(contigs):
    for contig in contigs:
        source_features = []

        # Scrape all source features
        for feature in contig.features:
            if feature.type != "source":
                continue

            # Ignore source features describing the whole contig
            if len(feature) >= len(contig.seq):
                continue

            source_features.append(feature)

        contig.features = source_features


def scrape_and_tabulate_host_metadata(contigs):
    """
    Put features of a prophage feature files into dictionaries.

    :param host_record_paths: list of files with prophage features
    :type file: list
    :return: characteristics of prophages
    :rtype: list
    """
    host_metadata = []
    for contig in contigs:
        keys = ["Prophage Name", "Left Coordinate", "Right Coordinate",
                "Length"]
        fwd_info = {key: list() for key in keys}
        rvs_info = {key: list() for key in keys}

        for feature in contig.features:
            if feature.location.strand == -1:
                info = rvs_info
            else:
                info = fwd_info
            # add the name
            info.get("Prophage Name").append(
                feature.qualifiers.get("locus_tag")[0])
            # add the location Left Coordinates
            info.get("Left Coordinate").append(feature.location.nofuzzy_start)
            # add Location Right Coordinates
            info.get("Right Coordinate").append(feature.location.nofuzzy_end)
            # add length
            info.get("Length").append(
                feature.location.nofuzzy_end -
                feature.location.nofuzzy_start)

        fwd_table = pandas.DataFrame(fwd_info)
        rvs_table = pandas.DataFrame(rvs_info)

        if len(fwd_table) > 0:
            fwd_table_html = pretty_html_table.build_table(
                                            fwd_table, "green_light",
                                            font_size="14px",
                                            text_align="center",
                                            font_family=DEFAULT_FONT_FAMILY)
        else:
            fwd_table_html = ""

        if len(rvs_table) > 0:
            rvs_table_html = pretty_html_table.build_table(
                                            rvs_table, "red_light",
                                            font_size="14px",
                                            text_align="center",
                                            font_family=DEFAULT_FONT_FAMILY)
        else:
            rvs_table_html = ""

        table_html = "\n".join([fwd_table_html, rvs_table_html])
        host_metadata.append(table_html)

    return host_metadata


def create_host_graphic_resources(contigs, resources_dir):
    host_graphic_resources = []
    for contig in contigs:
        host_translator = CircularSourceFeatureTranslator()

        graphic_record = host_translator.translate_record(
                                            contig,
                                            record_class=CircularGraphicRecord)

        ax, _ = graphic_record.plot(figure_width=10, with_ruler=True)

        # Write center label
        ax.text(0, -1, contig.id,
                horizontalalignment="center", verticalalignment="center",
                fontfamily="monospace", fontsize=30, fontweight="demibold")

        # Write center bp
        ax.text(0, -1.1,
                " ".join([str(len(contig)), "bp"]),
                horizontalalignment="center", verticalalignment="center",
                fontfamily="monospace", fontsize=12, fontweight="normal")

        resource_filepath = resources_dir.joinpath(contig.id).with_suffix(
                                                                        ".svg")
        ax.figure.savefig(resource_filepath, bbox_inches="tight")
        pyplot.close(ax.figure)

        filehandle = open(resource_filepath, "r")

        lines = filehandle.readlines()
        lines = "".join(lines)
        host_graphic_resources.append(lines)

        filehandle.close()

    return host_graphic_resources


def embed_prophage_genomes(prophages):
    translator = LinearFeatureTranslator()

    embedded_prophage_plots = []
    for prophage in prophages:
        graphic_record = translator.translate_record(prophage.record)
        bokeh_plot = plot_with_bokeh(graphic_record, figure_width=100,
                                     figure_height="auto", tools="auto",
                                     label_link_width=0,
                                     label_text_font=DEFAULT_FONT_FAMILY)

        # Format primary title
        bokeh_plot.title.text = prophage.id
        bokeh_plot.title.align = "left"
        bokeh_plot.title.text_font = DEFAULT_FONT_FAMILY
        bokeh_plot.title.text_font_size = "18px"

        # Add secondary location title
        location_prefix = ""
        if prophage.strand == -1:
            location_prefix = "complement"

        location = "".join([location_prefix, "(", str(prophage.start),
                            "..", str(prophage.end), ")"])
        bokeh_plot.add_layout(Title(text=location, align="left",
                                    text_font_size="12px",
                                    text_font=DEFAULT_FONT_FAMILY), "above")

        embedded_plot = file_html(bokeh_plot, CDN, prophage.id)
        embedded_prophage_plots.append(embedded_plot)

    return embedded_prophage_plots


# BOKEH HELPER OVERRIDE
# -----------------------------------------------------------------------------

# If you're wondering why there's a redundant function here,
# ask why dna_features_viewer.GraphicRecord.plot_with_bokeh hard codes so much
def plot_with_bokeh(graphic_record, figure_width=100, figure_height="auto",
                    title=None, tools="auto", feature_box_color="#000000",
                    label_link_color="#000000", label_text_font="arial",
                    label_text_font_size="12px",
                    label_text_font_style="normal", label_link_width=0.5):
    """Plots the graphic record using Bokeh"""
    if tools == "auto":
        tools = "xpan,xwheel_zoom,reset,tap"

    # Scrape matplotlib plot info

    ax, (features_levels, plot_data) = graphic_record.plot(
                                                figure_width=figure_width,
                                                annotate_inline=False)
    width, height = [int(100 * e) for e in ax.figure.get_size_inches()]
    pyplot.close(ax.figure)

    if figure_height == "auto":
        height = int(0.5 * height)
    else:
        height = 100 * figure_height

    # Ensures that height of the plot is 185 or greater, the minimal height
    # to see all icons
    height = max(height, 185)

    max_y = max(
                [data["annotation_y"] for f, data in plot_data.items()]
                + list(features_levels.values()))

    # Build the Bokeh plot
    plot = plotting.figure(plot_width=width, plot_height=height, tools=tools,
                           x_range=Range1d(0, graphic_record.sequence_length),
                           y_range=Range1d(-1, max_y + 1))

    # Define different hovertools
    cds_tooltips = [("gene", "@gene"), ("product", "@product"),
                    ("location", "@location")]
    cds_hovertool = HoverTool(tooltips=cds_tooltips, mode="mouse",
                              names=["cds"])
    plot.add_tools(cds_hovertool)

    att_tooltips = [("name", "@name"), ("sequence", "@sequence")]
    att_hovertool = HoverTool(tooltips=att_tooltips, mode="mouse",
                              names=["att"])
    plot.add_tools(att_hovertool)

    trna_tooltips = [("product", "@product"), ("codon", "@codon"),
                     ("location", "@location")]
    trna_hovertool = HoverTool(tooltips=trna_tooltips, mode="mouse",
                               names=["tRNA", "tmRNA"])
    plot.add_tools(trna_hovertool)

    # Set up patches source
    cds_feature_patches = list()
    att_feature_patches = list()
    trna_feature_patches = list()
    for feature, level in features_levels.items():
        bokeh_feature_patch = graphic_record.bokeh_feature_patch(
                                    feature.start, feature.end, feature.strand,
                                    figure_width=figure_width, level=level,
                                    color=feature.color, label=feature.label)

        location_prefix = ""
        if feature.strand == -1:
            location_prefix = "complement"

        location = "".join([location_prefix, "(", str(feature.start),
                            "..", str(feature.end), ")"])

        html_data = feature.html
        if html_data["gb_type"] == "CDS":
            bokeh_feature_patch["gene"] = html_data["gene"]
            bokeh_feature_patch["product"] = html_data["product"]
            bokeh_feature_patch["location"] = location
            cds_feature_patches.append(bokeh_feature_patch)
        elif feature.html["gb_type"] == "misc_recomb":
            bokeh_feature_patch["name"] = html_data["name"]
            bokeh_feature_patch["sequence"] = html_data["sequence"]
            att_feature_patches.append(bokeh_feature_patch)
        elif feature.html["gb_type"] in ("tRNA", "tmRNA"):
            bokeh_feature_patch["product"] = html_data["product"]
            bokeh_feature_patch["codon"] = html_data["codon"]
            bokeh_feature_patch["location"] = location
            trna_feature_patches.append(bokeh_feature_patch)

    if cds_feature_patches:
        cds_patches_source = plotting.ColumnDataSource(
                                            pandas.DataFrame.from_records(
                                                    cds_feature_patches))
        plot.patches(xs="xs", ys="ys", color="color", name="cds",
                     line_color=feature_box_color, source=cds_patches_source)

    if att_feature_patches:
        att_patches_source = plotting.ColumnDataSource(
                                            pandas.DataFrame.from_records(
                                                    att_feature_patches))
        plot.patches(xs="xs", ys="ys", color="color", name="att",
                     line_color=feature_box_color, source=att_patches_source)

    if trna_feature_patches:
        trna_patches_source = plotting.ColumnDataSource(
                                            pandas.DataFrame.from_records(
                                                    trna_feature_patches))
        plot.patches(xs="xs", ys="ys", color="color", name="tRNA",
                     line_color=feature_box_color, source=trna_patches_source)

    # Set up floating label and connector source
    bokeh_floating_labels = list()
    bokeh_label_segments = list()
    for feature, pdata in plot_data.items():
        bokeh_floating_label = dict(
                                x=feature.x_center, y=pdata["annotation_y"],
                                text=feature.label, color=feature.color)
        bokeh_floating_labels.append(bokeh_floating_label)

        bokeh_segment = dict(x0=feature.x_center, x1=feature.x_center,
                             y0=pdata["annotation_y"], y1=pdata["feature_y"])
        bokeh_label_segments.append(bokeh_segment)

    if plot_data != {}:
        labels_source = plotting.ColumnDataSource(
                                        pandas.DataFrame.from_records(
                                                    bokeh_floating_labels))
        plot.text(x="x", y="y", text="text", text_align="center",
                  text_font_size=label_text_font_size,
                  text_font=value(label_text_font),
                  text_font_style=label_text_font_style,
                  source=labels_source)

        segments_source = plotting.ColumnDataSource(
                                        pandas.DataFrame.from_records(
                                                    bokeh_label_segments))
        plot.segment(x0="x0", x1="x1", y0="y0", y1="y1",
                     line_width=label_link_width, color=label_link_color,
                     source=segments_source)

    plot.yaxis.visible = False
    plot.outline_line_color = None
    plot.grid.grid_line_color = None
    plot.toolbar.logo = None

    return plot



import shutil

from bitarray import bitarray, util as bit_util

from depht.data import GLOBAL_VARIABLES, PARAMETERS
from depht.functions.fasta import parse_fasta, write_fasta
from depht.functions.subprocess import run_command

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
WORKING_DIR_NAME = GLOBAL_VARIABLES["shell_db"]["dir_name"]

ENDIANESS = "big"
FASTA_PATH_NAME = "genes"


# PHAMERATION DEFAULTS
# -----------------------------------------------------------------------------
CMODE = PARAMETERS["phameration"]["cmode"]
CSTEP = PARAMETERS["phameration"]["cstep"]
SENS = PARAMETERS["phameration"]["sens"]
IDENT = PARAMETERS["phameration"]["ident"]
COVER = PARAMETERS["phameration"]["cover"]
EVALUE = PARAMETERS["phameration"]["evalue"]

MIN_GCS = PARAMETERS["phameration"]["min_gcs"]


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------
def assemble_bacterial_mask(contigs, bacterial_fasta, gene_bit_value_path,
                            working_dir):
    """Creates a binary mask for annotated coding regions in the input
    sequence contigs by using mmseqs2 clustering to identify bacterial-looking
    genes.

    :param contigs: Input sequence contig seqrecords
    :type contigs: list[SeqRecord]
    :param bacterial_fasta: Path to a fasta of reference bacterial genes
    :type bacterial_fasta: pathlib.Path
    :param gene_bit_value_path: Path to a hexadecimal clade representation file
    :type gene_bit_value_path: pathlib.Path
    :param working_dir: Path to place created data files.
    :type working_dir: pathlib.Path
    :return: A binary bacterial mask for each inputted sequence contig
    :rtype: list[list[int]]
    """
    # Parse hexadecimal clade representation file
    bacterial_gene_bit_values, clades = parse_gene_bit_value_file(
                                                        gene_bit_value_path)

    # Create working file
    fasta_path = working_dir.joinpath(FASTA_PATH_NAME).with_suffix(".fasta")

    # Write and index input genes and initialize bacterial mask
    bacterial_masks, gene_bit_values = initialize_bacterial_mask(
                                         contigs, bacterial_fasta, fasta_path)

    # Cluster genes
    clustering_map = cluster_bacterial_genes(fasta_path, working_dir)

    # Assign a clade representation bit array to each input gene
    assign_gene_bit_values(clustering_map, bacterial_gene_bit_values,
                           gene_bit_values, bacterial_masks, clades)

    # Create a gene bit array mask from the value of the most represented clade
    clade_bit_mask = assign_clade(gene_bit_values)

    # Mark input genes well represented in the assigned clade
    mark_bacterial_mask(bacterial_masks, gene_bit_values,
                        clade_bit_mask)

    # Write bacterial masks to file
    dump_bacterial_masks(contigs, bacterial_masks, clade_bit_mask, working_dir)

    # Store bacterial mask within each contig object
    for i in range(len(contigs)):
        contig = contigs[i]
        mask = bacterial_masks[i]

        contig.update_mask_bits(mask)

    # Cleanup working file
    fasta_path.unlink()

    return bacterial_masks


def initialize_bacterial_mask(contigs, bacterial_fasta, out_path):
    """Initialize a binary bacterial mask, index, and write each gene
    in the inputted sequence contigs.

    :param contigs:  Input sequence contig seqrecords
    :type contigs: list[SeqRecord]
    :param b_gene_names: Reference bacterial gene names
    :type b_gene_names: list[str]
    :param b_gene_seqs: Reference bacterial gene sequences
    """
    # Parse reference bacterial gene multiple-sequence fasta
    b_gene_names, b_gene_seqs = parse_fasta(bacterial_fasta)
    b_gene_names = list(range(len(b_gene_names)))

    # Initialize bacterial mask data structures
    bacterial_masks = []
    gene_bit_values = []
    for contig_index, contig in enumerate(contigs):
        feature_index = 0
        for feature in contig.genes:
            # Index input gene
            b_gene_names.append(f"{contig_index}_{feature_index}")

            # Add input gene translation
            b_gene_seqs.append(str(feature.qualifiers["translation"][0]))

            feature_index += 1

        # Initialize default values for each input gene
        bacterial_masks.append([1] * feature_index)
        gene_bit_values.append([None] * feature_index)

    # Write bacterial reference genes and input genome genes to file
    write_fasta(b_gene_names, b_gene_seqs, out_path)

    return bacterial_masks, gene_bit_values


def cluster_bacterial_genes(fasta_path, working_dir):
    """Clusters input genes with MMseqs2.

    :param fasta_path: Path to a fasta with input and reference genes.
    :type fasta_path: pathlib.Path
    :param working_dir: Path to place created data files
    :type working_dir: pathlib.Path
    """
    # Create database file directory
    database_dir = working_dir.joinpath("database")
    database_dir.mkdir(exist_ok=True)

    # Define mmseqs working file paths
    sequence_db = database_dir.joinpath("sequenceDB")
    cluster_db = database_dir.joinpath("clusterDB")
    seqfile_db = database_dir.joinpath("seqfileDB")
    result_file = database_dir.joinpath("clustering_results.txt")
    tmp_dir = database_dir.joinpath("tmp")
    tmp_dir.mkdir(exist_ok=True)

    # Create the initial mmseqs DB
    mmseqs_createdb(fasta_path, sequence_db)

    # Cluster bacterial genes in linear time
    mmseqs_linclust(sequence_db, cluster_db, tmp_dir, CMODE,
                    IDENT, COVER, EVALUE)

    # Create a database from the clustering
    mmseqs_createseqfiledb(sequence_db, cluster_db, seqfile_db)

    # Write the clustering results to file
    mmseqs_result2flat(sequence_db, sequence_db, seqfile_db,
                       result_file)

    # Parse the clustering output
    clustering_results = parse_mmseqs(result_file)

    # Remove database file directory
    shutil.rmtree(database_dir)

    return clustering_results


def assign_gene_bit_values(clustering_map, bacterial_gene_bit_values,
                           gene_bit_values, bacterial_masks, clades):
    """Assigns genes values contained in bit arrays depending on clade
    representation of reference bacterial genes.

    :param clustering_map: Cluster ID to cluster gene-member map
    :type clustering_map: dict
    :param bacterial_gene_bit_values: Bit arrays for bacterial reference genes.
    :type bacterial_gene_bit_values: list
    :param gene_bit_values: Initialized gene bit value list
    :type gene_bit_values: list
    :param bacterial_masks: Initialized bacterial core gene bits
    :type bacterial_masks: list
    :param clades: Number of total clades
    :type clades: int
    """
    for cluster, gene_names in clustering_map.items():
        gois = []
        cluster_bitarray = None
        absolute_bitarray = None
        for gene_name in gene_names:
            # Identify and store indexed input sequence genes
            if "_" in gene_name:
                gois.append(gene_name)
                continue

            # Reidentify bacterial reference gene index
            gene_index = int(gene_name)

            # Initialize / OR cluster member bitarray(s)
            # given reference gene values, to determine clade representation
            if cluster_bitarray is None:
                cluster_bitarray = bacterial_gene_bit_values[gene_index]
            else:
                # Ensure input bitarrays are the same length
                cluster_bitarray, b_gene_bitarray = equalize_bitarrays(
                                        cluster_bitarray,
                                        bacterial_gene_bit_values[gene_index])

                # OR reference gene bitarrays
                cluster_bitarray = (cluster_bitarray | b_gene_bitarray)

            # Intialize / AND cluster member bitarrays()
            # given reference gene values, to determine high confidence
            # bacterial genes
            if absolute_bitarray is None:
                absolute_bitarray = bacterial_gene_bit_values[gene_index]
            else:
                absolute_bitarray, b_gene_bitarray = equalize_bitarrays(
                                        absolute_bitarray,
                                        bacterial_gene_bit_values[gene_index])

                absolute_bitarray = (absolute_bitarray & b_gene_bitarray)

        for goi in gois:
            goi_split = goi.split("_")

            # Retrieve contig and gene indicies from input gene name
            contig_i = int(goi_split[0])
            gene_i = int(goi_split[1])

            # Assign input gene a value based on it's reference relatives
            gene_bit_values[contig_i][gene_i] = cluster_bitarray

            if absolute_bitarray is not None:
                if absolute_bitarray.count() >= clades:
                    bacterial_masks[contig_i][gene_i] = 0


def assign_clade(gene_bit_values, min_gcs=MIN_GCS):
    """Assigns clade membership of the input sequences based on the majority
    gene bit value

    :param gene_bit_values: Bitarrays of the clade representation for each gene
    :type gene_bit_values: list[bitarray]
    :return: A bitarray mask for the majority-assigned clade
    :rtype: bitarray
    """
    bit_count = list()
    for contig_gene_bitarrays in gene_bit_values:
        for gene_bitarray in contig_gene_bitarrays:
            if gene_bitarray is None:
                continue

            for index in range(len(gene_bitarray)):
                bit = gene_bitarray[index]
                if index >= len(bit_count):
                    bit_count.append(0)

                if bit:
                    bit_count[index] += 1

    if len(bit_count) <= 0:
        return bitarray([0])

    count = max(bit_count)
    if (count / len(gene_bit_values)) < min_gcs:
        return bitarray([0] * len(bit_count), endian=ENDIANESS)

    clade = bit_count.index(max(bit_count))

    clade_bit_mask = bitarray([0] * len(bit_count), endian=ENDIANESS)
    clade_bit_mask[clade] = 1

    return clade_bit_mask


def mark_bacterial_mask(bacterial_masks, gene_bit_values, clade_bit_mask):
    """Marks genes as bacterial with the given clade value bit mask.

    :param bacterial_masks: Initialized binary bacterial mask list
    :type bacterial_masks: list[list[int]]
    :param gene_bit_values: Bitarrays of the clade representation for each gene
    :type gene_bit_values: list[bitarray]
    :param clade_bit_mask: Bitarray mask for the majority-asssigned clade
    :type clade_bit_mask: bitarray
    """
    for contig_i, bacterial_mask in enumerate(bacterial_masks):
        for gene_i, bacterial_bit in enumerate(bacterial_mask):
            if bacterial_bit == 0:
                continue

            gene_bitarray = gene_bit_values[contig_i][gene_i]

            if gene_bitarray is None:
                continue

            masked_bitarray = (gene_bitarray & clade_bit_mask)

            if masked_bitarray.count() > 0:
                bacterial_mask[gene_i] = 0


def dump_bacterial_masks(contigs, bacterial_masks, clade_bit_mask,
                         working_dir):
    for i, contig in enumerate(contigs):
        bacterial_mask = bacterial_masks[i]
        filepath = working_dir.joinpath(".".join([contig.id, "txt"]))

        filehandle = filepath.open(mode="w")

        for bit in clade_bit_mask:
            filehandle.write(str(int(bit)))

        filehandle.write("\n")

        for bit in bacterial_mask:
            filehandle.write(str(int(bit)))


# HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def equalize_bitarrays(a_bitarray, b_bitarray):
    length_disparity = len(a_bitarray) - len(b_bitarray)

    if length_disparity < 0:
        for _ in range(abs(length_disparity)):
            a_bitarray.append(0)
    elif length_disparity > 0:
        for _ in range(length_disparity):
            b_bitarray.append(0)

    return a_bitarray, b_bitarray


# MMSEQS HELPER FUNCTIONS
# -----------------------------------------------------------------------------

def mmseqs_createdb(fasta, mmseqsdb):
    """
    Creates an MMseqs2 database at `mmseqsdb` using `fasta` as its input.

    :param fasta: path to the FASTA file to assemble into an MMseqs2 database
    :type fasta: pathlib.Path
    :param mmseqsdb: path to the desired MMseqs2 database
    :type mmseqsdb: pathlib.Path
    :return:
    """
    c = f"mmseqs createdb {fasta} {mmseqsdb} -v 3"
    run_command(c)


def mmseqs_cluster(sequence_db, cluster_db, tmp_dir, clustermode, clustersteps,
                   sensitivity, minseqid, coverage, evalue):
    """
    Clusters an MMseqs2 sequence database using the given parameters.

    :param sequence_db: path to the MMseqs2 sequence database to cluster
    :type sequence_db: pathlib.Path
    :param cluster_db: path to the MMseqs2 output cluster database
    :type cluster_db: pathlib.Path
    :param tmp_dir: temporary directory that MMseqs2 can use
    :type tmp_dir: pathlib.Path
    :param clustermode: mmseqs cluster --cluster-mode
    :type clustermode: int
    :param clustersteps: mmseqs cluster --cluster-steps
    :type clustersteps: int
    :param sensitivity: mmseqs cluster -s
    :type sensitivity: float
    :param minseqid: mmseqs cluster --min-seq-id
    :type minseqid: float
    :param coverage: mmseqs cluster -c
    :type coverage: float
    :param evalue: mmseqs cluster -e
    :type evalue: float
    :return:
    """
    c = (f"mmseqs cluster {str(sequence_db)} {str(cluster_db)} {str(tmp_dir)} "
         f"-v 3 --max-seqs 1000 --cluster-mode {clustermode} "
         f"--cluster-steps {clustersteps} -s {sensitivity} "
         f"--min-seq-id {minseqid} -c {coverage} -e {evalue}")
    run_command(c)


def mmseqs_linclust(sequence_db, cluster_db, tmp_dir, clustermode,
                    minseqid, coverage, evalue):
    """
    Clusters an MMseqs2 sequence database using the given parameters in
    linear time.

    :param sequence_db: path to the MMseqs2 sequence database to cluster
    :type sequence_db: pathlib.Path
    :param cluster_db: path to the MMseqs2 output cluster database
    :type cluster_db: pathlib.Path
    :param tmp_dir: temporary directory that MMseqs2 can use
    :type tmp_dir: pathlib.Path
    :param clustermode: mmseqs cluster --cluster-mode
    :type clustermode: int
    :param minseqid: mmseqs cluster --min-seq-id
    :type minseqid: float
    :param coverage: mmseqs cluster -c
    :type coverage: float
    :param evalue: mmseqs cluster -e
    :type evalue: float
    :return:
    """
    c = (f"mmseqs linclust {sequence_db} {cluster_db} {tmp_dir} "
         f"-v 3 --cluster-mode {clustermode} "
         f"--min-seq-id {minseqid} -c {coverage} -e {evalue}")
    run_command(c)


def mmseqs_result2profile(sequence_db, cluster_db, profile_db):
    """
    Converts an MMseqs2 cluster output database to a profile database.

    :param sequence_db: path to the clustered MMseqs2 sequence database
    :type sequence_db: pathlib.Path
    :param cluster_db: path to the MMseqs2 cluster database
    :type cluster_db: pathlib.Path
    :param profile_db: path to the desired MMseqs2 profile database
    :type profile_db: pathlib.Path
    :return:
    """
    c = (f"mmseqs result2profile {str(sequence_db)} {str(sequence_db)} "
         f"{str(cluster_db)} {str(profile_db)} -v 3")
    run_command(c)


def mmseqs_profile2consensus(profile_db, consensus_db):
    """
    Extracts consensus sequences from an MMseqs2 profile database and creates
    an MMseqs2 sequence database from those consensus sequences.

    :param profile_db: path to the MMseqs2 profile database
    :type profile_db: pathlib.Path
    :param consensus_db: path to a MMseqs2 consensus sequence database
    :type consensus_db: pathlib.Path
    :return:
    """
    c = f"mmseqs profile2consensus {profile_db} {consensus_db} -v 3"
    run_command(c)


def mmseqs_search(profile_db, consensus_db, align_db, tmp_dir, minseqid,
                  coverage, evalue):
    """
    Searches an MMseqs2 profile database against an MMseqs2 profile consensus
    sequence database, for HMM-based clustering.

    :param profile_db: path to the MMseqs2 profile database
    :type profile_db: pathlib.Path
    :param consensus_db: path to the MMseqs2 consensus sequence database
    :type consensus_db: pathlib.Path
    :param align_db: path to the desired MMseqs2 search result database
    :type align_db: pathlib.Path
    :param tmp_dir: temporary directory that MMseqs2 can use
    :type tmp_dir: pathlib.Path
    :param minseqid: mmseqs search --min-seq-id
    :type minseqid: float
    :param coverage: mmseqs search -c
    :type coverage: float
    :param evalue: mmseqs search --e-profile
    :type evalue: float
    :return:
    """
    c = (f"mmseqs search {profile_db} {consensus_db} {align_db} {tmp_dir} "
         f"-v 3 --max-seqs 1000 --min-seq-id {minseqid} -c {coverage} "
         f"--cov {coverage} -e {evalue} --e-profile {evalue} "
         "--add-self-matches")
    run_command(c)


def mmseqs_clust(consensus_db, align_db, result_db):
    """
    Clusters the MMseqs2 result database from running 'mmseqs search'.

    :param consensus_db: path to the MMseqs2 consensus sequence database
    :type consensus_db: pathlib.Path
    :param align_db: path to the MMseqs2 search result database
    :type align_db: pathlib.Path
    :param result_db: path to the desired MMseqs2 clustered result database
    :type result_db: pathlib.Path
    :return:
    """
    c = f"mmseqs clust {consensus_db} {align_db} {result_db} -v 3"
    run_command(c)


def mmseqs_createseqfiledb(sequence_db, cluster_db, sf_db):
    """
    Creates an MMseqs2 seqfileDB from a clustering output database.

    :param sequence_db: the path to the query db in MMseqs2 clustering
    :type sequence_db: pathlib.Path
    :param cluster_db: the path to the MMseqs2 clustering result database
    :type cluster_db: pathlib.Path
    :param sf_db: the path to the desired seqfileDB
    :type sf_db: pathlib.Path
    :return:
    """
    c = f"mmseqs createseqfiledb {sequence_db} {cluster_db} {sf_db} -v 3"
    run_command(c)


def mmseqs_result2flat(query_db, subject_db, result_db, output):
    """
    Converts an MMseqs2 result seqfileDB to a FASTA-like output file.

    :param query_db: the database used as the query in MMseqs2 clustering
    :type query_db: pathlib.Path
    :param subject_db: the database used as the subject in MMseqs2 clustering
    :type subject_db: pathlib.Path
    :param result_db: the MMseqs2 clustering result database
    :type result_db: pathlib.Path
    :param output: the desired clustering output file
    :type output: pathlib.Path
    :return:
    """
    c = f"mmseqs result2flat {query_db} {subject_db} {result_db} {output} -v 3"
    run_command(c)


# FILEIO
# -----------------------------------------------------------------------------
def parse_mmseqs(filepath):
    """
    Parses the indicated MMseqs2 output into a dictionary mapping
    phamids to the geneids that were put in them

    :param filepath: path to the MMseqs2 output file to parse
    :type filepath: pathlib.Path
    :return: phams
    """
    phams = dict()
    phamid = 0
    pham_genes = list()

    with filepath.open("r") as fh:
        prior = fh.readline()
        current = fh.readline()

        # While loop to iterate until EOF
        while current:
            if current.startswith(">"):
                # If current & prior both header lines - new pham block
                if prior.startswith(">"):
                    try:
                        pham_genes.pop(-1)
                    except IndexError:
                        pass
                    phams[phamid] = pham_genes
                    phamid += 1
                    pham_genes = list()
                pham_genes.append(current.lstrip(">").rstrip())
            else:
                # Do nothing on translation lines
                pass
            prior, current = current, fh.readline()
        # Need to dump the last p into the dictionary
        phams[phamid] = pham_genes
    phams.pop(0)  # 0th pham is placeholder

    return phams


def parse_gene_bit_value_file(filepath):
    """
    Parses the gene bit value file for determining input genome clade
    membership

    :param filepath: path to the gene bit value file
    :type filepath: pathlib.Path
    :return: Returns a list of bit arrays
    :rtype: list[bitarray]
    """
    bacterial_gene_bit_values = []
    filehandle = open(filepath, "rb")
    line = filehandle.readline()

    hex_values = line.split(b"_")

    clades = 0
    for hex_value in hex_values:
        if hex_value == b"":
            continue

        b_gene_bitarray = bit_util.hex2ba(hex_value, endian=ENDIANESS)

        bitarray_count = b_gene_bitarray.count()
        if bitarray_count > clades:
            clades = bitarray_count

        bacterial_gene_bit_values.append(b_gene_bitarray)

    filehandle.close()

    return bacterial_gene_bit_values, clades



import math

from depht.data import GLOBAL_VARIABLES, PARAMETERS
from depht.functions.blastn import blastn, REF_BLASTN_OUTFMT
from depht.functions.fasta import write_fasta
from depht.functions.statistics import transform, zscore

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
DEFAULT_PRODUCT = GLOBAL_VARIABLES["sequences"]["default_product"]

KMER_SIZE = PARAMETERS["att_detection"]["kmer_size"]
MIN_ATT_SCORE = PARAMETERS["att_detection"]["min_att_score"]
EVALUE_FILTER = PARAMETERS["att_detection"]["evalue_threshold"]
BLAST_SORT_KEY = PARAMETERS["att_detection"]["blast_sort_key"]

AQ_WEIGHT = PARAMETERS["att_detection"]["att_quality_weight"]
IP_WEIGHT = PARAMETERS["att_detection"]["integrase_proximity_weight"]
MC_WEIGHT = PARAMETERS["att_detection"]["model_coverage_weight"]
TR_WEIGHT = PARAMETERS["att_detection"]["trna_weight"]
RC_WEIGHT = PARAMETERS["att_detection"]["reference_concurrence_weight"]

L_SEQ_NAME = "putative_attL_region"
R_SEQ_NAME = "putative_attR_region"


def find_attachment_site(prophage, l_seq, r_seq,
                         reference_db_path, tmp_dir, sort_key=BLAST_SORT_KEY,
                         k=KMER_SIZE, min_score=MIN_ATT_SCORE,
                         l_name=L_SEQ_NAME, r_name=R_SEQ_NAME):
    """Given the sequences of a putative attL region and putative attR region,
    find the most probable attachment site, dictated by the sequence's length
    and it's distance from the predicted origin position.

    :param prophage: Prophage object to find an attachment site for
    :type prophage: depht.classes.prophage.Prophage
    :param l_seq: The sequence of a putative attL region.
    :type l_seq: str
    :param r_seq: The sequence of a putative attR region.
    :type r_seq: str
    :param tmp_dir: The directory for which to write sequences and files to
    :type tmp_dir: pathlib.Path
    :param k: Length of the word size storerd in the DeBruijn graph.
    :type k: int
    :param l_name: Name to give to the putative attL region sequence.
    :type l_name: str
    :param r_name: Name to give to the putative attR region sequence.
    :type r_name: str
    :return: A tuple of information associated with the detected att site.
    :rtype: tuple
    """
    # Write the putative attL region to file
    l_seq_path = tmp_dir.joinpath(f"{l_name}.fasta")
    write_fasta([l_name], [l_seq], l_seq_path)

    # Write the putative attR region to file
    r_seq_path = tmp_dir.joinpath(f"{r_name}.fasta")
    write_fasta([r_name], [r_seq], r_seq_path)
    # Calculate and store right region's coordinate start for easy access
    r_seq_start = prophage.end - len(r_seq)

    # Use BLASTn to retrieve putative attachment site sequences
    kmer_contigs = blast_attachment_site(l_seq_path, r_seq_path, tmp_dir, k=k,
                                         evalue=len(l_seq))

    if not kmer_contigs:
        return

    transform_kmer_contig_bitscores(kmer_contigs)

    paired_ref_map = find_reference_att_sites(l_seq_path, r_seq_path,
                                              reference_db_path, tmp_dir,
                                              k, sort_key,
                                              prophage.start, r_seq_start)

    # Score putative attachment site sequences
    # TODO paramaterize and allow access to change the 'exponent' variable
    scored_kmer_contigs = []
    for kmer_contig in kmer_contigs:
        scores = score_kmer(kmer_contig, prophage, paired_ref_map, r_seq_start)

        scored_kmer_contigs.append((kmer_contig, scores))

    # Sort attachment site sequences by score
    scored_kmer_contigs.sort(key=lambda x: x[1][0], reverse=True)

    att_table_path = tmp_dir.joinpath("att.txt")
    dump_attachment_sites(prophage, scored_kmer_contigs, att_table_path,
                          r_seq_start)

    attB_table_path = tmp_dir.joinpath("attB.txt")
    dump_reference_attB_sites(paired_ref_map, attB_table_path)

    kmer_contig, scores = scored_kmer_contigs[0]

    if scores[0] < min_score:
        return

    new_start = prophage.start + kmer_contig[1]
    new_end = r_seq_start + kmer_contig[2]

    return new_start, new_end, scores, kmer_contig[0]


# MAIN HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def find_reference_att_sites(left_seq_path, right_seq_path, reference_db_path,
                             tmp_dir, k, sort_key, l_seq_start, r_seq_start):
    # BLASTn both regions against the reference database
    left_map = build_reference_map(left_seq_path, reference_db_path, tmp_dir)
    right_map = build_reference_map(right_seq_path, reference_db_path, tmp_dir)

    ref_ids = list(set(left_map.keys()).intersection(set(right_map.keys())))

    paired_ref_map = pair_reference_maps(ref_ids, left_map, right_map,
                                         k, sort_key, l_seq_start, r_seq_start)

    return paired_ref_map


def blast_attachment_site(l_seq_path, r_seq_path, tmp_dir, k=KMER_SIZE,
                          evalue=EVALUE_FILTER):
    """Given the path to files containink the putative attL region and
    putative attR region, BLASTn the sequences of both regions against each
    other and retrieve matching sequences and their positions.

    :param l_seq_path: path to the sequence of the putative attL region
    :type l_seq_path: pathlib.Path
    :param r_seq_path: path to the sequence of the putative attR region
    :type r_seq_path: pathlib.Path
    :param tmp_dir: path where temporary files can go
    :type tmp_dir: pathlib.Path
    :param k: Length of the word size used by the BLAST algorithm
    :type k: int
    :return: A list of contigs and their positions in the sequence and graph.
    :rtype: list(tuple(str, int, int))
    """
    blast_results = blastn(
        l_seq_path, r_seq_path, tmp_dir, mode="subject", word_size=k,
        evalue=evalue, gapopen=10, gapextend=4)

    kmer_contigs = []
    for result in blast_results:
        # Append the contig and its positions to the list
        kmer_contig = [result["qseq"],
                       int(result["qstart"]) - 1, int(result["send"]),
                       float(result["bitscore"])]
        kmer_contigs.append(kmer_contig)

    return kmer_contigs


def dump_attachment_sites(prophage, scored_kmer_contigs, outpath, r_seq_start):
    filehandle = outpath.open(mode="w")

    for kmer_contig, scores in scored_kmer_contigs:
        new_start = prophage.start + kmer_contig[1]
        new_end = r_seq_start + kmer_contig[2]

        att_line_data = [new_start, new_end, len(kmer_contig[0])]
        score_line_data = [round(score, 2) for score in scores]
        seq_data = [kmer_contig[0]]

        line_data = att_line_data + score_line_data + seq_data
        line_data = [str(line_entry) for line_entry in line_data]

        filehandle.write("\t".join(line_data))
        filehandle.write("\n")

    filehandle.close()


def dump_reference_attB_sites(paired_ref_map, outpath):
    filehandle = outpath.open(mode="w")

    for ref_id, ref_data in paired_ref_map.items():
        line_data = [str(ref_data_entry)
                     for ref_data_entry in list(ref_data[4:9]) + [ref_data[3]]]

        filehandle.write("\t".join(line_data))
        filehandle.write("\n")

    filehandle.close()


# REFERENCE BLASTING HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def build_reference_map(sequence_path, reference_db_path, tmp_dir):
    """Maps sequence BLASTn aligned reference genome IDs to their respective
    alignment result data.

    :param sequence_path: Path to query to be aligned to the reference database
    :type sequence_path: pathlib.Path
    :param reference_db_path: Path to the database of references to search
    :type reference_db_path: pathlib.Path
    :param tmp_dir: Working directory to place BLASTn inputs and outputs
    :type tmp_dir: pathlib.Path
    :return: A map of aligned reference genome IDs to alignment result data
    """
    # Try to retrieve reference results for the sequence to the references
    blast_results = blastn(sequence_path, reference_db_path, tmp_dir,
                           outfmt=REF_BLASTN_OUTFMT)

    reference_map = dict()
    for blast_result in blast_results:
        # Checks to see if the sequence reference ID has already been stored
        results = reference_map.get(blast_result["sseqid"], list())

        results.append(blast_result)
        reference_map[blast_result["sseqid"]] = results

    return reference_map


def pair_reference_maps(ref_ids, left_map, right_map, k, sort_key,
                        l_seq_start, r_seq_start):
    paired_ref_map = {}
    for ref_id in ref_ids:
        ref_data = []

        for l_data in left_map[ref_id]:
            for r_data in right_map[ref_id]:
                # Find the coordinate ranges of the aligned reference genomes
                left_ref_range = range(int(l_data["sstart"]),
                                       int(l_data["send"]))
                right_ref_range = range(int(r_data["sstart"]),
                                        int(r_data["send"]))

                # Determine if the coordinate ranges overlap
                overlap_range = set(left_ref_range).intersection(
                                                        set(right_ref_range))
                overlap_len = len(overlap_range)

                # If the overlap length meets the minimum att length
                # treat the sequence as a putative attB that we can use
                # as a reference to set the boundaries for the prophage
                if overlap_len >= k:
                    # Find the right coordinate of the putative attL
                    # in the aligned left region of the prophage
                    l_qend = int(l_data["qend"])

                    # Find the left coordinate of the putative attR
                    # in the aligned right region of the prophage
                    r_qstart = int(r_data["qstart"])

                    new_start = (l_seq_start + int(l_qend) - overlap_len)
                    new_end = (r_seq_start + int(r_qstart) + overlap_len)

                    score = float(l_data[sort_key]) + float(r_data[sort_key])
                    att_data = (new_start, new_end, overlap_len, score,
                                ref_id,
                                l_data["sstart"], l_data["send"],
                                r_data["sstart"], r_data["send"])

                    ref_data.append(att_data)

            ref_data.sort(key=lambda x: x[3], reverse=True)

            if ref_data:
                paired_ref_map[ref_id] = ref_data[0]

    return paired_ref_map


# SCORING HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def transform_kmer_contig_bitscores(kmer_contigs):
    """Transform the kmer contig bitscores into z-scores
    :param kmer_contigs: A list of kmers and their positions in the sequence
    :type kmer_contigs: list
    """
    bitscores = []
    for kmer_contig in kmer_contigs:
        bitscores.append(kmer_contig[3])

    zscores = zscore(bitscores)
    transform(zscores, min_t=0, max_t=1)

    for i, kmer_contig in enumerate(kmer_contigs):
        kmer_contig.append(zscores[i])


def score_kmer(kmer_contig, prophage, paired_ref_map, r_seq_start):
    """Score kmer contigs with a complete holsitic approach.

    :param kmer_contig: A tuple containing the contig, and its positions
    :type kmer_contig: tuple(str, int, int)
    :return: The score of the given kmer
    :rtype: float
    """
    attL_pos = prophage.start + kmer_contig[1]
    attR_pos = r_seq_start + kmer_contig[2]

    att_quality_score = score_att_quality(kmer_contig[4])

    int_proximity_score, int_distance = score_integrase_proximity(
                                                    prophage,
                                                    attL_pos - prophage.start,
                                                    attR_pos - prophage.start)

    model_cov_score, model_coverage = score_model_coverage(attR_pos - attL_pos,
                                                           len(prophage.seq))

    reference_score, reference_bitscore = score_reference_concurrence(
                                        attL_pos, attR_pos,
                                        len(kmer_contig[0]), paired_ref_map)

    composite_score = ((att_quality_score + int_proximity_score +
                        model_cov_score) *
                       (3 / (AQ_WEIGHT + IP_WEIGHT + MC_WEIGHT)) +
                       reference_score)

    return (composite_score, att_quality_score, kmer_contig[3],
            int_proximity_score, int_distance,
            model_cov_score, model_coverage,
            reference_score, reference_bitscore)


def score_att_quality(normalized_bitscore, weight=AQ_WEIGHT):
    score = normalized_bitscore

    if score < 0:
        score = 0.0
    elif score > 1:
        score = 1.0

    weighted_score = score * weight
    return weighted_score


def score_integrase_proximity(prophage, attL_pos, attR_pos, base_dist=1500,
                              weight=IP_WEIGHT):
    int_dist = None
    for feature in prophage.record.features:
        if feature.type != "CDS":
            continue

        product = feature.qualifiers.get("product", [DEFAULT_PRODUCT])[0]

        if "integrase" in product or "transposase" in product:
            left_int_dist = int(feature.location.start - attL_pos)
            right_int_dist = int(attR_pos - feature.location.end)

            if int_dist is None:
                int_dist = left_int_dist
            elif int_dist < 0 and left_int_dist > int_dist:
                int_dist = left_int_dist
            else:
                if left_int_dist < int_dist and left_int_dist > 0:
                    int_dist = left_int_dist

            if int_dist is None:
                int_dist = right_int_dist
            elif int_dist < 0 and right_int_dist > int_dist:
                int_dist = right_int_dist
            else:
                if right_int_dist < int_dist and right_int_dist > 0:
                    int_dist = right_int_dist

    if int_dist is None:
        score = -1
        int_dist = -1
    elif int_dist < 0:
        score = -1
    elif int_dist == 0:
        score = 1
    else:
        score = (base_dist / int_dist) ** 2

        if score > 1:
            score = 1

    weighted_score = score * weight
    return weighted_score, int_dist


def score_trna_overlap(prophage, attL_pos, attR_pos, att_len,
                       weight=TR_WEIGHT):
    overlap = 0
    for feature in prophage.record.features:
        if feature.type != "tRNA":
            continue

        trna_range = set(range(feature.location.start, feature.location.end))

        attL_range = set(range(attL_pos, attL_pos + att_len))
        attR_range = set(range(attR_pos - att_len, attR_pos))

        if trna_range.intersection(attL_range):
            overlap = 1
        elif trna_range.intersection(attR_range):
            overlap = 1

        if overlap > 0:
            break

    if overlap <= 0:
        return (0, overlap)
    else:
        return (overlap * TR_WEIGHT, overlap)


def score_model_coverage(putative_len, model_len, weight=MC_WEIGHT):
    score = putative_len / model_len

    if score > 1:
        score = 1.0

    weighted_score = score * weight
    return score, weighted_score


def score_reference_concurrence(attL_pos, attR_pos, att_len, paired_ref_map,
                                base=10, weight=RC_WEIGHT):
    """Return a score dependant on the coordinate range overlap of the
    kmer contig with a preducted reference attB.
    """
    ref_bitscore = None
    for ref_id, ref_data in paired_ref_map.items():
        attL_range = set(range(attL_pos, (attL_pos + att_len)))
        ref_attL_range = set(range(ref_data[0], (ref_data[0] + att_len)))

        if not attL_range.intersection(ref_attL_range):
            continue

        attR_range = set(range((attR_pos - att_len), attR_pos))
        ref_attR_range = set(range((ref_data[1] - att_len), ref_data[1]))

        if not attR_range.intersection(ref_attR_range):
            continue

        if ref_bitscore is None:
            ref_bitscore = ref_data[3]
        else:
            if ref_data[3] > ref_bitscore:
                ref_bitscore = ref_data[3]

    if ref_bitscore is None:
        return 0.0, 0

    if ref_bitscore < 10:
        ref_bitscore = 10

    score = 1 - (1 / math.log(ref_bitscore, base))

    if score < 0:
        score = 0.0
    elif score > 1:
        score = 1.0

    weighted_score = score * weight
    return weighted_score, ref_bitscore



"""Utility script to make a blastn database that DEPhT can use for
attB searches.
"""
import argparse
import pathlib
import sys

from Bio import SeqIO

from depht.data import GLOBAL_VARIABLES
from depht_train.functions import blastdb

NAME = GLOBAL_VARIABLES["reference_db"]["name"]


# MAIN FUNCTIONS
#  ----------------------------------------------------------------------------
def parse_args(unparsed_args):
    """Parse commandline arguments.

    :param unparsed_args:
    :type unparsed_args: list[str]
    """
    parser = argparse.ArgumentParser(description=__doc__)

    parser.add_argument("input_dir", type=pathlib.Path)
    parser.add_argument("output_dir", type=pathlib.Path)

    parser.add_argument("-v", "--verbose", action="store_true")
    parser.add_argument("-n", "--name", type=str, default=NAME)

    args = parser.parse_args(unparsed_args)
    return args


def build_reference_db(input_dir, output_dir, name=NAME, verbose=False):
    """

    :param input_dir:
    :param output_dir:
    :param name:
    :param verbose:
    :return:
    """
    output_dir.mkdir(exist_ok=True, parents=True)

    cc_fasta_path = output_dir.joinpath(f"{name}.fasta")
    write_concatenated_fasta(input_dir, cc_fasta_path)

    blastdb.create_blastdb(cc_fasta_path, output_dir, name, verbose=verbose)


def write_concatenated_fasta(input_dir, cc_fasta_path):
    with cc_fasta_path.open(mode="w") as filehandle:
        for genome_file in input_dir.iterdir():
            if not genome_file.is_file():
                continue

            genome_records = SeqIO.parse(genome_file, "fasta")
            for genome_record in genome_records:
                filehandle.write(f">{genome_record.id}\n")
                filehandle.write(f"{str(genome_record.seq)}\n")


def main(unparsed_args=None):
    """Commandline entrypoint for this module."""
    if not unparsed_args:
        unparsed_args = sys.argv

    if len(unparsed_args) == 1:
        unparsed_args.append("-h")

    args = parse_args(unparsed_args[1:])
    build_reference_db(args.input_dir, args.output_dir, name=args.name,
                       verbose=args.verbose)


if __name__ == "__main__":
    main()



"""Class for easy visualization of progress level for multiprocessing
or multithreading."""


class ProgressBar:
    """Inline-updatable progress bar."""
    def __init__(self, current, end, width=50):
        self._percent = int(float(current) / end * 100)
        self._width = int(width)
        self._ratio = int(100 / width)
        self._multi = int(self._percent / self._ratio)
        self._pad = self._width - self._multi

    def show(self):
        """Print ProgressBar to the console."""
        if self._percent < 100:
            print("\r" + str(self), end="")
        else:
            print("\r" + str(self))

    def __str__(self):
        s = f"[{'#' * self._multi}{' ' * self._pad}] {self._percent}%"
        return s


def show_progress(current, end, width=50):
    """Create an instance of ProgressBar and print it in-line.

    :param current: current step (1 through n)
    :type current: int
    :param end: number of steps (n)
    :type end: int
    :param width: character width for the progressbar
    :type width: int
    :return: progressbar
    """
    ProgressBar(current, end, width).show()



import json
import os
import shlex
import shutil
from subprocess import DEVNULL, PIPE, Popen


# CREATE DATABASE
# -----------------------------------------------------------------------------
def create_hhsuitedb(aln_dir, db_dir, db_name, cores=1, use_mpi=False,
                     verbose=False, versions=None):
    """Function to create a database compatible with the HHsuite3 toolkit.

    :param aln_dir: Path to a directory containing fasta-formatted MSA files.
    :type aln_dir: pathlib.Path
    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param db_name: Name of the database to be created.
    :type db_name: str
    :param cores: Number of subprocesses to spawn during database construction.
    :type cores: int
    :param use_mpi: Toggles use of the mpirun library for parallel-processing.
    :type use_mpi: bool
    """
    if verbose:
        stdout = PIPE
    else:
        stdout = DEVNULL

    if verbose:
        print("Creating multiple sequence ffindex file...")
    msa_fftuple = build_msa_ffindex(aln_dir, db_dir, db_name, stdout=stdout)

    # A hhsuite database requires pairs of three main file types:
    # A .ffdata and .ffindex pair containing a3m formatted MSAs
    if verbose:
        print("Condensing multiple sequence ffindex file into a3m file...")
    a3m_fftuple = convert_a3m_ffindex(db_dir, db_name, msa_fftuple,
                                      cores=cores, stdout=stdout,
                                      use_mpi=use_mpi)

    # A .ffdata and .ffindex pair containing formatted HHsuite3 HMMs
    if verbose:
        print("Creating Hidden Markov Model ffindex file...")
    hmm_fftuple = convert_hmm_ffindex(db_dir, db_name, a3m_fftuple,
                                      cores=cores, stdout=stdout,
                                      use_mpi=use_mpi)

    # A .ffdata and .ffindex pair containing cs219 sequences for prefiltering
    if verbose:
        print("Creating cs219 ffindex file...")
    cs219_ffindex = create_cs219_ffindex(db_dir, db_name, stdout=stdout)

    # The MSA and HMM ffindex files are sorted according to data from cs219
    # These steps are essential to the speed of the HHsuite3 utilities
    if verbose:
        print("Sorting ffindex files for fast database indexing...")
    sorting_file = create_sorting_file(cs219_ffindex, db_dir, stdout=stdout)
    sort_ffindex(db_dir, sorting_file, a3m_fftuple, stdout=stdout)
    sort_ffindex(db_dir, sorting_file, hmm_fftuple, stdout=stdout)

    # Removes msa_ffindex files, as they are not needed for a functional db
    msa_fftuple[0].unlink()
    msa_fftuple[1].unlink()
    sorting_file.unlink()

    # Is a placeholder.
    # Requires additional work for the functionality to be realized.
    if not verify_hhsuite_database(db_dir, db_name):
        print(f"Inconsistencies detected in HHsuite database "
              f"at '{db_dir}'.\n  Scrapping database.")
        shutil.rmtree(db_dir)

    # Creates a json file that dumps key: value pairs with no file extension
    # Is unnecessesary, only for convenience and easy tab completion
    db_path = db_dir.joinpath(db_name)
    if versions is not None:
        create_version_file(db_path, versions)

    return db_path


def build_msa_ffindex(aln_dir, db_dir, db_name, stdout=None):
    """Function to build a ffindex database pair from a directory of
    fasta-formatted MSA files

    :param aln_dir: Path to a directory containing fasta-formatted MSA files.
    :type aln_dir: pathlib.Path
    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param db_name: Name of the database to be created.
    :type db_name: str
    :param stdout: Standard output to use for the spawned process
    :returns: A tuple containing path objects for the created ffindex database
    :rtype: tuple(pathlib.Path, pathlib.Path)
    """
    if stdout is None:
        stdout = DEVNULL

    msa_ffdata = db_dir.joinpath("".join([db_name, "_msa.ffdata"]))
    msa_ffindex = db_dir.joinpath("".join([db_name, "_msa.ffindex"]))

    command = f"ffindex_build -s '{msa_ffdata}' '{msa_ffindex}' '{aln_dir}'"

    split_command = shlex.split(command)
    with Popen(args=split_command, stdout=stdout, stderr=stdout) as process:
        out, errors = process.communicate()

    return (msa_ffdata, msa_ffindex)


def build_hmm_ffindex(hmm_dir, db_dir, db_name, stdout=None):
    """Function to build a ffindex database pair from a directory of
    HHsuite3 formatted HMM files

    :param hmm_dir: Path to a directory containing HHsuite3 HMM files.
    :type aln_dir: pathlib.Path
    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param db_name: Name of the database to be created.
    :type db_name: str
    :param stdout: Standard output to use for the spawned process
    :returns: A tuple containing path objects for the created ffindex database
    :rtype: tuple(pathlib.Path, pathlib.Path)
    """
    if stdout is None:
        stdout = DEVNULL

    hmm_ffdata = db_dir.joinpath("".join([db_name, "_hmm.ffdata"]))
    hmm_ffindex = db_dir.joinpath("".join([db_name, "_hmm.ffindex"]))

    command = f"ffindex_build -s '{hmm_ffdata}' '{hmm_ffindex}' '{hmm_dir}'"

    split_command = shlex.split(command)
    with Popen(args=split_command, stdout=stdout, stderr=stdout) as process:
        out, errors = process.communicate()

    return (hmm_ffdata, hmm_ffindex)


def convert_a3m_ffindex(db_dir, db_name, msa_fftuple, cores=1, stdout=None,
                        use_mpi=False):
    """Function to convert a fasta-formatted MSA ffindex database into an
    a3m formatted ffindex database

    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param db_name: Name of the database to be created.
    :type db_name: str
    :param msa_fftuple: A tuple containing path objects to an MSA ffindex db.
    :type msa_fftuple: tuple(pathlib.Path, pathlib.Path)
    :param cores: Number of subprocesses to spawn during database construction.
    :type cores: int
    :param stdout: Standard output to use for the spawned process(es)
    :param use_mpi: Toggles use of the mpirun library for parallel-processing.
    :type use_mpi: bool
    :returns: A tuple containing path objects for the created ffindex database.
    :rtype: tuple(pathlib.Path, pathlib.Path)
    """
    if stdout is None:
        stdout = DEVNULL

    a3m_ffdata = db_dir.joinpath("".join([db_name, "_a3m.ffdata"]))
    a3m_ffindex = db_dir.joinpath("".join([db_name, "_a3m.ffindex"]))

    # This command assumes that the msa_fftuple holds the path to the ffindex
    # data file at index 0 and the ffindex index file at index 1.
    # The parameter set is supplied by the HHsuite3 documentation.
    command = (f"'{msa_fftuple[0]}' '{msa_fftuple[1]}' "
               f"-d '{a3m_ffdata}' -i '{a3m_ffindex}' "
               "-- hhconsensus -M 50 -maxres 65535 "
               "-i stdin -oa3m stdout -v 0")

    # If parallel processing with mpirun is desired, use ffindex_apply_mpi.
    # ffindex_apply_mpi is tricky to install due to dependancies, be careful.
    if use_mpi:
        os.environ["OMP_NUM_THREADS"] = "1"
        command = "".join([(f"mpirun -np {cores} ffindex_apply_mpi "),
                           command])
    else:
        command = "".join(["ffindex_apply ", command])

    split_command = shlex.split(command)
    with Popen(args=split_command, stdout=stdout, stderr=stdout) as process:
        out, errors = process.communicate()

    return (a3m_ffdata, a3m_ffindex)


def convert_hmm_ffindex(db_dir, db_name, a3m_fftuple, cores=1, stdout=None,
                        use_mpi=False):
    """Function to convert a fasta-formatted MSA ffindex database into an
    HHsuite3 formatted HMM ffindex database

    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param db_name: Name of the database to be created.
    :type db_name: str
    :param a3m_fftuple: A tuple containing path objects to an MSA ffindex db.
    :type a3m_fftuple: tuple(pathlib.Path, pathlib.Path)
    :param cores: Number of subprocesses to spawn during database construction.
    :type cores: int
    :param stdout: Standard output to use for the spawned process(es)
    :param use_mpi: Toggles use of the mpirun library for parallel-processing.
    :type use_mpi: bool
    :returns: A tuple containing path objects for the created ffindex database.
    :rtype: tuple(pathlib.Path, pathlib.Path)
    """
    if stdout is None:
        stdout = DEVNULL

    hmm_ffdata = db_dir.joinpath("".join([db_name, "_hmm.ffdata"]))
    hmm_ffindex = db_dir.joinpath("".join([db_name, "_hmm.ffindex"]))

    # This command assumes that the a3m_fftuple holds the path to the ffindex
    # data file at index 0 and the ffindex index file at index 1.
    command = (f"'{a3m_fftuple[0]}' '{a3m_fftuple[1]}' "
               f"-i '{hmm_ffindex}' -d '{hmm_ffdata}' "
               "-- hhmake -i stdin -o stdout -v 0")

    # If parallel processing with mpirun is desired, use ffindex_apply_mpi.
    # ffindex_apply_mpi is tricky to install due to dependancies, be careful.
    if use_mpi:
        os.environ["OMP_NUM_THREADS"] = "1"
        command = "".join([f"mpirun -np {cores} ffindex_apply_mpi ",
                           command])
    else:
        command = "".join(["ffindex_apply ", command])

    split_command = shlex.split(command)
    with Popen(args=split_command, stdout=stdout, stderr=stdout) as process:
        out, errors = process.communicate()

    return (hmm_ffdata, hmm_ffindex)


def create_cs219_ffindex(db_dir, db_name, cores=1, use_mpi=False,
                         stdout=None):
    """Function to convert a a3m formatted MSA ffindex database into
    a cs219 sequence database for fast prefiltering

    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param db_name: Name of the database to be created.
    :type db_name: str
    :param cores: Number of subprocesses to spawn during database construction.
    :type cores: int
    :param use_mpi: Toggles use of the mpirun library to parallel-processing.
    :type use_mpi: bool
    :param stdout: Standarad output to use for the spawned process(es)
    :returns: Path to the created cs219 sequence ffindex index file
    :rtype: pathlib.Path
    """
    if stdout is None:
        stdout = DEVNULL

    # Annoyingly, this function requires that the path given has no file ext.
    a3m_ffpath = db_dir.joinpath("".join([db_name, "_a3m"]))
    cs219_ffpath = db_dir.joinpath("".join([db_name, "_cs219"]))

    # The parameter set is supplied by the HHsuite3 documentation
    command = ("cstranslate -f -x 0.3 -c 4 -I a3m "
               f"-i '{a3m_ffpath}' -o '{cs219_ffpath}'")

    if use_mpi:
        command = "".join([f"mpirun -np {cores}", command])

    split_command = shlex.split(command)
    with Popen(args=split_command, stdout=stdout, stderr=stdout) as process:
        out, errors = process.communicate()

    cs219_ffindex = db_dir.joinpath("".join([db_name, "_cs219.ffindex"]))
    return cs219_ffindex


def create_sorting_file(cs219_ffindex, db_dir, sorting_file=None,
                        stdout=None):
    """Function to create a sorted index from a cs219 sequences file.

    :param cs219_ffindex: Path to a ffindex index file for cs219 sequences.
    :type cs219_ffindex: pathlib.Path
    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param sorting_file: Desired path to create the sorting file
    :type sorting_file: pathlib.Path
    :param stdout: Standard output to use for the spawned process
    :returns: Path to the created sorting file for sorting ffindex databases.
    :rtype: pathlib.Path
    """
    if sorting_file is None:
        sorting_file = db_dir.joinpath("sorting.dat")

    # The parameter set is supplied by the HHsuite3 documentation
    sort_command = f"sort -k3 -n -r '{cs219_ffindex}' "
    cut_command = "cut -f1"

    split_sort_command = shlex.split(sort_command)
    split_cut_command = shlex.split(cut_command)

    # Prepares to pipe the output of the sort command into cut
    # Cut grabs the first field and writes it out
    sort_process = Popen(args=split_sort_command, stdout=PIPE)
    cut_process = Popen(args=split_cut_command, stdout=PIPE,
                        stdin=sort_process.stdout)
    sort_process.stdout.close()

    # Writes the output of the cut command into file, after decoding
    with sorting_file.open(mode="w") as filehandle:
        out, err = cut_process.communicate()
        filehandle.write((out).decode("utf-8"))

    return sorting_file


def sort_ffindex(db_dir, sorting_file, ffindex_tuple, stdout=None):
    """Function to sort a ffindex database with a sorting key.

    :param db_dir: Path to a directory where the database will be created.
    :type db_dir: pathlib.Path
    :param sorting_file: Path to a sorting file key for ffindex databases.
    :type sorting_file: pathlib.Path
    :param ffindex_tuple: A tuple containing path objects to a ffindex db.
    :type ffindex_tuple: tuple(pathlib.Path, pathlib.Path)
    """
    if stdout is None:
        stdout = DEVNULL

    # Defines the path for the ordered ffindex database
    ord_ffdata = db_dir.joinpath("".join([ffindex_tuple[0].stem,
                                          "ordered.ffdata"]))
    ord_ffindex = db_dir.joinpath("".join([ffindex_tuple[1].stem,
                                           "ordered.ffindex"]))

    command = (f"ffindex_order '{sorting_file}' "
               f"'{ffindex_tuple[0]}' '{ffindex_tuple[1]}' "
               f"'{ord_ffdata}' '{ord_ffindex}'")

    split_command = shlex.split(command)
    with Popen(args=split_command, stdout=stdout, stderr=stdout) as process:
        out, errors = process.communicate()

    # Overwrites the unsorted ffindex database files with sorted files
    ord_ffdata.replace(ffindex_tuple[0])
    ord_ffindex.replace(ffindex_tuple[1])


# Placeholder function
def verify_hhsuite_database(db_dir, db_name, cores=1, stdout=None):
    if stdout is None:
        stdout = DEVNULL

    return True


# HELPER FUNCTIONS
# -----------------------------------------------------------------------------
def create_version_file(filepath, versions_dict):
    with filepath.open(mode="w") as filehandle:
        json.dump(versions_dict, filehandle)



import csv

from depht.functions.subprocess import run_command

# GLOBAL VARIABLES
# -----------------------------------------------------------------------------
BLASTN_OUTFMT = "10 sseqid qstart qend sstart send length gapopen " \
                "mismatch evalue bitscore qseq"

REF_BLASTN_OUTFMT = "10 sseqid qstart qend sstart send length gapopen " \
                    "mismatch evalue bitscore"
BLASTN_EVALUE = 1E-05


# MAIN FUNCTIONS
# -----------------------------------------------------------------------------
def blastn(query, target, tmp_dir, mode="db", evalue=BLASTN_EVALUE,
           word_size=None, gapopen=None, gapextend=None, outfmt=BLASTN_OUTFMT):
    """
    Runs blastn in either query/subject mode or query/database mode, as
    indicated by `mode`. Returns hits better than `evalue`.

    NOTE: **kwargs will be interpreted as additional blastn parameters.

    :param query: the query FASTA file to use
    :type query: pathlib.Path
    :param target: the target to BLAST against
    :type target: pathlib.Path
    :param tmp_dir: the directory where temporary files can go
    :type tmp_dir: pathlib.Path
    :param mode: how to treat the target (db or subject sequence)
    :type mode: str
    :param evalue: the e-value cutoff to use
    :type evalue: float
    :param word_size: specify a word size (>=4) to use with blastn
    :type word_size: int or None
    :param gapopen: specify a gap open penalty to use with blastn
    :type gapopen: int or None
    :return: results
    """
    # Store any results here
    results = list()

    # Create output filepath
    outfile = tmp_dir.joinpath(f"{query.stem}_blastn_results.csv")

    # Prepare blastn command
    if mode == "db":
        command = f"blastn -query {query} -db {target}"
    elif mode == "subject":
        command = f"blastn -query {query} -subject {target}"
    else:
        raise ValueError("valid blastn modes are: 'db', 'subject'")
    command += f" -out {outfile} -outfmt '{outfmt}'"

    if evalue:
        command += f" -evalue {evalue}"

    if word_size:
        command += f" -word_size {word_size}"

    if gapopen:
        command += f" -gapopen {gapopen}"

    # TODO:
    #  Gap extend and gap open are dependant on each other, for the future,
    #  there needs to be some logic mandating the use of either none or both
    if gapextend:
        command += f" -gapextend {gapextend}"

    run_command(command)

    # Return parsed hits as list of dictionaries
    fields = outfmt.split()[1:]
    try:
        blastn_reader = open(outfile, "r")
        csv_reader = csv.DictReader(blastn_reader, fieldnames=fields)
        for row in csv_reader:
            results.append(row)
        blastn_reader.close()
    except FileNotFoundError:
        pass

    return results


def locate_subsequence(query, subject, tmp_dir):
    """
    Uses `blastn` method to find the coordinates of a query sequence
     in a target sequence.

    :param query: the query sequence to search
    :type query: pathlib.Path
    :param subject: the subject sequence to search
    :type subject: pathlib.Path
    :param tmp_dir: the directory where temporary files can go
    :type tmp_dir: pathlib.Path
    :return: start, end
    """
    blast_results = blastn(query, subject, tmp_dir, mode="subject")

    # Return best alignment result location if any alignments were found
    if blast_results:
        start = int(blast_results[0]["sstart"])
        end = int(blast_results[0]["send"])
        return start, end



"""Helper script to facilitate downloading models from OSF.io."""

import argparse
import pathlib
import requests
import shutil
import sys
import zipfile

DEPHT_DIR = pathlib.Path().home().joinpath(".depht/models")

URLS = {"Gordonia": "https://osf.io/download/djwsb/",
        "Mycobacterium": "https://osf.io/download/aw4up/"}


def parse_args():
    """Parse command line arguments."""
    p = argparse.ArgumentParser(description=__doc__)

    p.add_argument("model",
                   type=str, choices=["Mycobacterium", "Gordonia"],
                   help="choose a model to download")

    p.add_argument("-f", "--force", action="store_true",
                   help="overwrite existing model files if they exist "
                        "[default: False]")
    p.add_argument("-v", "--verbose", action="store_true",
                   help="print verbose output to stdout [default: False]")

    return p.parse_args()


def download_model(url, path):
    """Download model from OSF.io and save to path."""
    r = requests.get(url, allow_redirects=True)
    open(path, "wb").write(r.content)

    return path


def unzip_model(path):
    """Unzip model to path."""
    with zipfile.ZipFile(path, "r") as zip_ref:
        zip_ref.extractall(path.parent)


def main():
    """Commandline entry point."""
    if len(sys.argv) == 1:
        sys.argv.append("-h")
    args = parse_args()

    model = args.model
    model_url = URLS[model]
    zip_path = DEPHT_DIR.joinpath(f"{model}.zip")
    model_path = DEPHT_DIR.joinpath(model)
    overwrite = args.force

    # Create the DEPhT data directory if it doesn't exist
    if not DEPHT_DIR.is_dir():
        DEPHT_DIR.mkdir(parents=True)

    # If model exists and not told to overwrite, exit
    if model_path.is_dir() and not overwrite:
        print(f"Model files already exist at {model_path}. "
              "Use '-f/--force' to overwrite.")
        sys.exit(1)

    # If model exists and told to overwrite, remove it
    if model_path.is_dir() and overwrite:
        if args.verbose:
            print(f"Removing existing model files at {model_path}...")
        shutil.rmtree(model_path)

    # Download and unzip model
    if args.verbose:
        print(f"Downloading model files from {model_url}...")
    download_model(model_url, zip_path)
    if args.verbose:
        print(f"Unzipping model files to {model_path}...")
    unzip_model(zip_path)
    if args.verbose:
        print(f"Removing zip file at {zip_path}...")
    zip_path.unlink()
    if args.verbose:
        print("Done.")


if __name__ == "__main__":
    main()



PARAMETERS = {
    "bacterial_sequences": {
        "phameration": {
            "first_iteration": {
                "--cluster-mode": 0,
                "--cluster-steps": 1,
                "-s": 8,
                "--min-seq-id": 0.50,
                "-c": 0.80,
                "-e": 0.001}},
        "pham_clust": {
            "gcs_threshold": 70}},
    "phage_sequences": {
        "phameration": {
            "first_iteration": {
                "--cluster-mode": 0,
                "--cluster-steps": 1,
                "-s": 8,
                "--min-seq-id": 0.45,
                "-c": 0.75,
                "-e": 0.001},
            "second_iteration": {
                "--min-seq-id": 0.3,
                "-c": 0.5,
                "-e": 0.001}},
        "pham_clust": {
            "gcs_threshold": 35}},
    "phage_homologs": {
        "annotation_consensus_threshold": 0.33,
        "min_hmm_count": 5,
        "essential_annotations": {
            "LIKE": ["terminase",
                     "major capsid",
                     "portal",
                     "lysin",
                     "integrase",
                     "immunity repressor",
                     "transposase",
                     "collar"],
            "NOT LIKE": ["hypothetical",
                         "putative",
                         "NKF"]},
        "extended_annotations": {
            "LIKE": [],
            "NOT LIKE": ["hypothetical",
                         "putative",
                         "NKF",
                         "teminase",
                         "lysin",
                         "immunity repressor",
                         "integrase",
                         "major capsid",
                         "portal",
                         "transposase",
                         "collar"]}},
    "shell_db": {"rep_threshold": 0.60},
    "reference_db": {},
    "classifier": {"window": 55}}
